/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.io.BufferedReader;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;

import javax.naming.NamingException;
import javax.xml.namespace.QName;
import javax.xml.rpc.Stub;

import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.URI;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.ietf.jgss.GSSCredential;
import org.oasis.wsn.SubscriptionManager;
import org.oasis.wsn.WSBaseNotificationServiceAddressingLocator;
import org.oasis.wsrf.faults.BaseFaultType;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.properties.ResourcePropertyValueChangeNotificationType;

import org.globus.axis.gsi.GSIConstants;
import org.globus.axis.util.Util;
import org.globus.exec.generated.FaultResourcePropertyType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.InternalStateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.ManagedExecutableJobResourcePropertiesType;
import org.globus.exec.generated.ManagedExecutableJobResourceStateType;
import org.globus.exec.generated.NameValuePairType;
import org.globus.exec.generated.ServiceLevelAgreementFaultType;
import org.globus.exec.generated.StateChangeNotificationMessageType;
import org.globus.exec.generated.StateChangeNotificationMessageWrapperType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.monitoring.JobStateMonitor;
import org.globus.exec.service.exec.utils.FileMapping;
import org.globus.exec.service.exec.utils.FileMappingException;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.factory.ManagedJobFactoryResource;
import org.globus.exec.service.job.ManagedJobResourceImpl;
import org.globus.exec.service.utils.DelegatedCredential;
import org.globus.exec.service.utils.GlobusShToolsProperties;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.service.ManagedJobHelper;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.DeleteType;
import org.globus.rft.generated.ReliableFileTransferPortType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.RFTFaultResourcePropertyType;
import org.globus.rft.generated.service.ReliableFileTransferServiceAddressingLocator;

import org.globus.security.gridmap.GridMap;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.util.GlobusURL;
import org.globus.util.I18n;

import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.Topic;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.core.notification.ResourcePropertyValueChangeNotificationElementType;
import org.globus.wsrf.impl.SimpleTopic;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;
import org.globus.wsrf.impl.security.authorization.SelfAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.security.SecureResource;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.PerformanceLog;

public class ManagedExecutableJobResource
    extends                                 ManagedJobResourceImpl
    implements                              RemoveCallback,
                                            SecureResource
{
    private static Log staticLogger =
        LogFactory.getLog(ManagedExecutableJobResource.class.getName());
    private Log logger;

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private Map variableMap;
    private static PerformanceLog performanceLogger = new PerformanceLog(
        ManagedExecutableJobResource.class.getName() + ".performance");

    JobStateMonitor jobMonitor = null;

    /**
     * System-specfic values to add to the script environment.
     */
    private String[] scriptEnvironment;

    /**
     * Script execution instance.
     */
    private JobManagerScript script;

    /**
     * RSL description of the job this object is managing.
     */
    private PerlJobDescription perlJobDescription;

    private static String STDIN_FILE_NAME = "stdin";

    private static int MAX_VARIABLE_DEPTH = 10;

    static
    {
        Util.registerTransport();
    }

    public void initializeWholeState(
        Object                                  wholeResourceState,
        QName                                   resourceElementQName,
        Object                                  key)
        throws                                  ResourceException
    {
        super.initializeWholeState(
            wholeResourceState, resourceElementQName, key);

        ManagedExecutableJobResourceStateType wholeJobState =
            (ManagedExecutableJobResourceStateType)wholeResourceState;
        setResourceData(wholeJobState.getResourceData());
        this.initialize(wholeJobState.getResourceProperties(),
                        resourceElementQName, key);
    }

    public void initialize(
            Object                          resourceBean,
            QName                           resourceElementQName,
            Object                          key)
            throws                          ResourceException
    {
        this.logger = LogFactory.getLog(
            ManagedExecutableJobResource.class.getName()
            + "." + ((ResourceKey)key).getValue().toString());

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("Entering initialize()");
        }

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("at super.initialize()");
        }
        super.initialize(resourceBean, resourceElementQName, key);

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("at initSecurity()");
        }
        initSecurity();

        JobDescriptionType jobDescription = getJobDescription();

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("at initVariableMap()");
        }
        initVariableMap();

        //set default directory if null
        String directory = jobDescription.getDirectory();
        if (directory == null) {
            jobDescription.setDirectory("${GLOBUS_USER_HOME}");
        }

        try
        {
            resolveVariables();
        }
        catch (FaultType f)
        {
            setFault(f);

            StateMachine.systemCancel(this);
        }
        catch (Exception e)
        {
            FaultType fault = FaultUtils.makeFault(
                FaultType.class, (ResourceKey) getID(),
                i18n.getMessage(Resources.VARIABLE_RESOLUTION_ERROR),
                e,
                getState(),
                null,
                0);
            setFault(fault);

            if (this.logger.isDebugEnabled()) {
                this.logger.trace("setting state and notifying");
            }

            StateMachine.systemCancel(this);
        }

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("at initPerlJobDescription()");
        }
        initPerlJobDescription();

        if (this.logger.isDebugEnabled()) {
            this.logger.trace("Perl Job Description: "
                        + this.perlJobDescription.toPerlString());
        }

        initVariableMap();

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("Leaving initialize()");
        }
    }

    protected void initializeTopics() {
        super.initializeTopics();
        Topic stateChangeTopic = new SimpleTopic(ManagedJobConstants.RP_STATE);
        this.getTopicList().addTopic(stateChangeTopic);
        //requires manual notifications in relevant parts of the code
    }

    public void remove()
        throws                              ResourceException
    {
        performanceLogger.start();

        InternalStateEnumeration internalState = null;
        StateEnumeration state = null;
        this.logger.trace("acquring lock on resource...");
        synchronized (this)
        {
            this.logger.trace("acquired lock on resource...");

            internalState = getInternalState();
            state = getState();

            //prevent multiple remove() calls from having any effect
            boolean canceled = ((Boolean) getResourceDatum(
                ManagedExecutableJobConstants.RD_CANCELED)).booleanValue();
            if (   !internalState.equals(InternalStateEnumeration.Done)
                && !state.equals(InternalStateEnumeration.Done)
                && !internalState.equals(InternalStateEnumeration.Failed)
                && !state.equals(InternalStateEnumeration.Failed)
                && !canceled)
            {
                StateMachine.userCancel(this);
            }

            if (this.logger.isDebugEnabled())
            {
                this.logger.debug("Remove called"
                            + " with external state " + state
                            + " and internal state " + internalState);
            }

            while (   !internalState.equals(InternalStateEnumeration.Done)
                   && !state.equals(InternalStateEnumeration.Done)
                   && !internalState.equals(InternalStateEnumeration.Failed)
                   && !state.equals(InternalStateEnumeration.Done))
            {
                this.logger.debug(
                    "Waiting to be Done or Failed. Current state: "
                        + internalState);
                try {
                    wait(2000); //wait two seconds for job to complete
                } catch (Exception e) {
                    String errorMessage =
                        i18n.getMessage(
                            Resources.INTERRUPTED_RESOURCE_REMOVE_WAIT);
                    this.logger.error(errorMessage, e);
                }
                internalState = getInternalState();
                state = getState();
            }
            this.logger.trace("releasing lock on resource...");
        }
        this.logger.trace("released lock on resource...");

        performanceLogger.stop("remove");
    }

    public void initSecurity() throws ResourceException
    {
        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("Entering initSecurity()");
        }

        //indirectly create user proxy and cache delegated credentials
        String localUserId
            = (String) getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID);
        File userGlobusDirectory = getUserGlobusDirectory(localUserId);

        //Create job proxy from delegated credential if delegation was done
        JobDescriptionType jobDescription = getJobDescription();
        if (jobDescription.getJobCredentialEndpoint() != null)
        {
            //performanceLogger.start();

            if (this.logger.isTraceEnabled())
            {
                this.logger.trace("at getJobCredential()");
            }
            try
            {
                getJobCredential(userGlobusDirectory);
            }
            catch (Exception e)
            {
                StateEnumeration state = getState();
                if (   !state.equals(StateEnumeration.Done)
                    && !state.equals(StateEnumeration.Failed))
                {
                    throw new ResourceException(e);
                }
            }

            //performanceLogger.stop("getJobCredential(userGlobusDirectory)");
        }

        //get security config
        ResourceSecurityConfig securityConfig= new ResourceSecurityConfig(
            ManagedJobConstants.SECURITY_CONFIG_FILE_NAME);
        try {
            securityConfig.init();
        } catch (Exception e) {
            throw new ResourceException(
                i18n.getMessage("resourceSecurityInitializationFailed"),
                e);
        }

        //get security descriptor
        this.securityDescriptor = securityConfig.getSecurityDescriptor();
        if (this.logger.isDebugEnabled()) {
            this.logger.trace("resource credential subject:\n"
                        + this.securityDescriptor.getSubject());
        }

        //Set the gridmap from resource properties set during resource creation
        String userSubject
            = (String) getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT);
        if (this.logger.isDebugEnabled()) {
            this.logger.trace("setting resource securty grid map..."
                        + "\nSubject: " + userSubject
                        + "\nLocal user ID: " + localUserId);
        }
        GridMap gridMap = new GridMap();
        gridMap.map(userSubject, localUserId);
        this.securityDescriptor.setGridMap(gridMap);

        if (this.logger.isTraceEnabled())
        {
            this.logger.trace("Leaving initSecurity()");
        }
    }

    public File getJobDirectory(
        String                              localUserId)
    {
        return new File(
            getUserGlobusDirectory(localUserId),
            ((ResourceKey)this.getID()).getValue().toString());
    }

    public static File getUserGlobusDirectory(
        String                              localUserId)
    {
        return new File(
            ManagedExecutableJobHome.getUserHomeDirectory(localUserId)
            + File.separator + ".globus");
    }

    private File getUserProxyFile(
        String                              localUserId)
        throws                              ResourceException
    {
        DelegatedCredential jobCredential = null;
        try
        {
            jobCredential
                = getJobCredential(getUserGlobusDirectory(localUserId));
        }
        catch (Exception e)
        {
            StateEnumeration state = getState();
            if (   !state.equals(StateEnumeration.Done)
                && !state.equals(StateEnumeration.Failed))
            {
                throw new ResourceException(e);
            }
        }
        if (jobCredential == null)
        {
            return null;
        }

        return jobCredential.getUserProxyFile();
    }

    protected boolean envVarAlreadySet(String varName)
    {
        JobDescriptionType jobDescription = getJobDescription();
        NameValuePairType[] envVars = jobDescription.getEnvironment();
        if (envVars == null)
        {
            return false;
        }

        boolean varFound = false;
        for (int index=0;(!varFound) && (index < envVars.length);index++)
        {
            if (envVars[index].getName().equals(varName))
            {
                varFound = true;
            }
        }

        return varFound;
    }

    private NameValuePairType[] initExtraEnvironment() throws ResourceException
    {
        String localUserId
            = (String) getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID);

        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String globusLocation = factoryHomeConfiguration.getGlobusLocation();
        String x509CertDir = factoryHomeConfiguration.getX509CertDir();
        String x509UserProxy = null;
        File userProxyFile = null;
        try
        {
            userProxyFile = getUserProxyFile(localUserId);
        }
         catch (ResourceException re)
        {
            StateEnumeration state = getState();
            if (   !state.equals(StateEnumeration.Done)
                && !state.equals(StateEnumeration.Failed))
            {
                throw re;
            }
        }
        if (userProxyFile == null)
        {
            x509UserProxy = "";
        } else
        {
            x509UserProxy = userProxyFile.toString();
        }
        String tcpPortRange = factoryHomeConfiguration.getTcpPortRange();

        //init script environment
        int index=0;
        String userHome
            = ManagedExecutableJobHome.getUserHomeDirectory(localUserId);
        this.scriptEnvironment = new String[6];

        this.scriptEnvironment[index++]
            = new String("HOME=" + userHome);
        this.scriptEnvironment[index++]
            = new String("GLOBUS_LOCATION=" + globusLocation);
        this.scriptEnvironment[index++]
            = new String("X509_CERT_DIR=" + x509CertDir);
        this.scriptEnvironment[index++]
            = new String("X509_USER_PROXY=" + x509UserProxy);
        this.scriptEnvironment[index++]
            = new String("X509_USER_CERT=" + x509UserProxy);
        this.scriptEnvironment[index++]
            = new String("X509_USER_KEY=" + x509UserProxy);

        //init proc environment
        Vector extraProcEnvironment = new Vector();
        index = 0;
        JobDescriptionType jobDescription = getJobDescription();

        try {
            if (!envVarAlreadySet("GLOBUS_LOCATION"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "GLOBUS_LOCATION",
                    globusLocation));
            }
            if (!envVarAlreadySet("X509_CERT_DIR"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "X509_CERT_DIR",
                    x509CertDir));
            }
            if (!envVarAlreadySet("X509_USER_PROXY"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "X509_USER_PROXY",
                    x509UserProxy));
            }
            if (!envVarAlreadySet("X509_USER_CERT"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "X509_USER_CERT",
                    x509UserProxy));
            }
            if (!envVarAlreadySet("X509_USER_KEY"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "X509_USER_KEY",
                    x509UserProxy));
            }
            if (!envVarAlreadySet("HOME"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "HOME",
                    resolveVariableInString(
                        "environment",
                        "${GLOBUS_USER_HOME}")));
            }
            if (!envVarAlreadySet("LOGNAME"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "LOGNAME",
                    resolveVariableInString(
                        "environment",
                        "${GLOBUS_USER_NAME}")));
            }
            if (!envVarAlreadySet("GLOBUS_TCP_PORT_RANGE"))
            {
                if (tcpPortRange != null && !tcpPortRange.equals("")) {
                    extraProcEnvironment.add(new NameValuePairType(
                        "GLOBUS_TCP_PORT_RANGE", tcpPortRange));
                }
            }
            if (!envVarAlreadySet("SCRATCH_DIRECTORY"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "SCRATCH_DIRECTORY",
                    resolveVariableInString(
                        "environment",
                        "${GLOBUS_SCRATCH_DIR}")));
            }
            //for Java jobs
            if (!envVarAlreadySet("JAVA_HOME"))
            {
                extraProcEnvironment.add(new NameValuePairType(
                    "JAVA_HOME",
                    System.getProperty("java.home", "")));
            }

            //Rendezvous
            extraProcEnvironment.add(new NameValuePairType(
                "GLOBUS_GRAM_JOB_HANDLE",
                this.getResourceData().getHandle()));
        }
        catch (ServiceLevelAgreementFaultType slaf)
        {
            throw new ResourceException(slaf);
        }

        return (NameValuePairType[]) extraProcEnvironment.toArray(
            new NameValuePairType[0]);
    }

    private void initExtraAttributes()
    {
        if (this.logger.isDebugEnabled()) {
            this.logger.trace(
                "Adding extra attributes to the Perl job attribute map");
        }

        ManagedExecutableJobResourceDataType resourceData
            = (ManagedExecutableJobResourceDataType) getResourceData();

        String localUserId
            = (String) getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID);
        this.perlJobDescription.put(
            "jobdir",
            getJobDirectory(localUserId).toString());

        //useforkstarter
        String localResourceManager =
            ((ManagedExecutableJobResourceDataType)resourceData)
                .getLocalResourceManager();
        if (localResourceManager.equals(
            ManagedJobFactoryConstants.FACTORY_TYPE.FORK)) {
            this.perlJobDescription.put("useForkStarter", "1");
        }

        //xml_extensions
        this.perlJobDescription.put("xml_extensions", "1");

        //emitcondorprocesses
        this.logger.trace("checking for condorness of " + localResourceManager);
        if (localResourceManager.equals(
            ManagedJobFactoryConstants.FACTORY_TYPE.CONDOR)) {
            this.logger.trace("adding emitCondorProcesses attribute");
            this.perlJobDescription.put("emitCondorProcesses", "1");
        }

        ManagedJobFactoryHome factoryHome = null;
        ManagedJobFactoryResource factoryResource = null;
        try
        {
            factoryHome = ManagedJobFactoryHome.getHome();
            factoryResource = (ManagedJobFactoryResource) factoryHome.find(
                ManagedJobHelper.getResourceKey(localResourceManager));
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
        if (factoryResource.isDefaultSoftwareEnvironmentEnabled())
        {
            this.perlJobDescription.put(
                "enableDefaultSoftwareEnvironment",
                "1");
        }


        //jobid
        String[] localJobIds
            = ((ManagedExecutableJobResourceDataType)resourceData)
                .getLocalJobId();

        if (localJobIds != null) {
            int size = localJobIds.length;
            StringBuffer sb = new StringBuffer();

            for (int i = 0; i < size; i++) {
                sb.append((String) localJobIds[i]);

                if (i < size - 1) {
                    sb.append(",");
                }
            }

            this.perlJobDescription.put("jobid", sb.toString());
        }

        //condorOs
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String condorOs = factoryHomeConfiguration.getCondorOS();
        this.perlJobDescription.put("condorOs", condorOs);

        //condorArch
        String condorArch = factoryHomeConfiguration.getCondorArchitecture();
        this.perlJobDescription.put("condorArch", condorArch);
    }

    private void initVariableMap() throws ResourceException
    {
        variableMap = new HashMap();

        String localUserId
            = (String) getResourceProperty(
                ManagedJobConstants.RP_LOCAL_USER_ID);

        variableMap.put("GLOBUS_USER_NAME", localUserId);
        variableMap.put(
            "GLOBUS_USER_HOME",
            ManagedExecutableJobHome.getUserHomeDirectory(localUserId));

        String localResourceManager
            = ((ManagedExecutableJobResourceDataType)getResourceData())
                .getLocalResourceManager();
        ManagedJobFactoryHome factoryHome = null;
        try {
            factoryHome = ManagedJobFactoryHome.getHome();
        } catch (NamingException ne) {
            throw new ResourceException(ne);
        }
        ManagedJobFactoryResource factoryResource =
                (ManagedJobFactoryResource) factoryHome.find(
                        ManagedJobHelper.getResourceKey(localResourceManager));
        String scratchDir
            = (String) factoryResource.getResourcePropertySet().get(
                ManagedJobFactoryConstants.RP_SCRATCH_BASE_DIRECTORY).get(0);
        if (this.logger.isDebugEnabled()) {
            this.logger.trace("GLOBUS_SCRATCH_DIR:" + scratchDir);
        }
        variableMap.put("GLOBUS_SCRATCH_DIR", scratchDir);

        variableMap.put("GLOBUS_JOB_ID", ((ResourceKey)getID()).getValue());

        //create mappings for all values in the substitutionDefinitionsFile
        Iterator subDefIterator
            = factoryResource.substitutionDefinitionSet().iterator();
        while (subDefIterator.hasNext())
        {
            Map.Entry subDef = (Map.Entry)subDefIterator.next();
            variableMap.put(subDef.getKey(), subDef.getValue());
        }
    }

    private void initPerlJobDescription()
    {
        JobDescriptionType jobDescription = getJobDescription();
        try {
            NameValuePairType[] extraEnvironment = initExtraEnvironment();

            this.perlJobDescription = new PerlJobDescription(
                jobDescription, extraEnvironment);
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                "objectCreationFailed",
                new Object[]{PerlJobDescription.class.getName()});
            throw new RuntimeException(errorMessage, e);
        }

        initExtraAttributes();
    }

    private static String cleanUrl(String urlString)
        throws                              MalformedURLException
    {
        GlobusURL url = new GlobusURL(urlString);

        //check for an invalid file URL
        String protocol = url.getProtocol();
        String host = url.getHost();
        if (protocol.equals("file") && (host.length() > 0))
        {
            throw new MalformedURLException(i18n.getMessage(
                Resources.UNALLOWED_FILE_URL_HOST,
                new String[] { url.getURL(), url.getHost() }));
        }
        else if (!protocol.equals("file") && (host.length() == 0))
        {
            throw new MalformedURLException(i18n.getMessage(
                Resources.MISSING_URL_HOST,
                new String[] { url.getURL() }));
        }

        //normalize the path
        String cleanedPath = url.getPath();
        if (cleanedPath != null)
        {
            cleanedPath
                = cleanedPath.replaceAll("/+", "/").replaceFirst("^/", "");
        }

        int port = url.getPort();

        String cleanedUrlString
            = url.getProtocol() + "://"
            + url.getHost()
            + (port>=0?":"+port:"") + "/"
            + (cleanedPath!=null?cleanedPath:"");
        staticLogger.debug("Cleaned URL: " + cleanedUrlString);

        return cleanedUrlString;
    }

    public void resolveVariables()
        throws                              ServiceLevelAgreementFaultType,
                                            MalformedURLException,
                                            FileMappingException
    {
        this.logger.debug("Entering resolveVariables()");

        JobDescriptionType jobDescription = getJobDescription();
        String str;
        String [] arr;
        NameValuePairType [] env;
        String res;

        FileMapping fsMap = null;

        //fileStageIn
        TransferRequestType stageIn = jobDescription.getFileStageIn();
        if (stageIn != null)
        {
            if (fsMap == null)
            {
                fsMap = getFileSystemMapping();
            }
            TransferType[] stageInList = stageIn.getTransfer();
            for (int index=0; index<stageInList.length; index++)
            {
                String sourceUrl = null;
                String destinationUrl = null;
                try {
                    sourceUrl
                        = cleanUrl(stageInList[index].getSourceUrl());
                } catch (Exception e)
                {
                    this.logger.debug("unable to clean URL \""
                                + stageInList[index].getSourceUrl()
                                + "\"",
                                e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageIn",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }
                try {
                    destinationUrl
                        = cleanUrl(stageInList[index].getDestinationUrl());
                } catch (Exception e)
                {
                    this.logger.debug("unable to clean URL \""
                                + stageInList[index].getDestinationUrl()
                                +"\"",
                                e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageIn",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }

                //sourceUrl
                if (this.logger.isTraceEnabled()) {
                    this.logger.trace("fileStageIn sourceUrl before: "
                                +sourceUrl);
                }
                sourceUrl = resolveVariableInString(
                    "fileStageIn", sourceUrl);
                try {
                    GlobusURL url = new GlobusURL(sourceUrl);
                    if (url.getProtocol().equals("file")) {
                        sourceUrl = null;
                    }
                } catch (java.net.MalformedURLException mue) {
                    this.logger.warn(mue);
                    sourceUrl = null;
                }

                if (sourceUrl == null) {
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createStagingFault(
                            "null",
                            destinationUrl,
                            "fileStageIn",
                            getState(),
                            (String) null,
                            (ResourceKey) getID());
                    this.logger.debug("sourceUrl fault", fault);

                    throw fault;
                }

                try {
                    stageInList[index].setSourceUrl(cleanUrl(sourceUrl));
                } catch (Exception e)
                {
                    this.logger.debug(
                        "unable to clean URL \"" + sourceUrl + "\"", e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageIn",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }
                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileStageIn sourceUrl after:"
                                +sourceUrl);
                }
                //END sourceUrl

                //destinationUrl
                if (this.logger.isTraceEnabled()) {
                    this.logger.trace("fileStageIn destinationUrl before: "
                                +destinationUrl);
                }
                destinationUrl = resolveVariableInString(
                    "fileStageIn", destinationUrl);
                GlobusURL url = new GlobusURL(destinationUrl);
                if (url.getProtocol().equals("file")) {
                    destinationUrl = fsMap.mapFileName(url);
                }

                if (destinationUrl == null) {
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createStagingFault(
                            sourceUrl,
                            "null",
                            "fileStageIn",
                            getState(),
                            (String) null,
                            (ResourceKey) getID());
                    this.logger.debug("sourceUrl fault", fault);

                    throw fault;
                }

                try {
                    stageInList[index].setDestinationUrl(
                        cleanUrl(destinationUrl));
                } catch (Exception e)
                {
                    this.logger.debug(
                        "unable to clean URL \"" + destinationUrl
                        + "\"", e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageIn",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                        throw fault;
                }
                if (this.logger.isTraceEnabled()) {
                    this.logger.trace("fileStageIn destinationUrl after: "
                                +destinationUrl);
                }
                //END destinationUrl
            }
        }
        //END fileStageIn

        //fileStageOut
        TransferRequestType stageOut = jobDescription.getFileStageOut();
        if (stageOut != null)
        {
            if (fsMap == null)
            {
                fsMap = getFileSystemMapping();
            }

            TransferType[] stageOutList = stageOut.getTransfer();
            for (int index=0; index<stageOutList.length; index++)
            {
                String sourceUrl = null;
                String destinationUrl = null;
                try {
                    sourceUrl
                        = cleanUrl(stageOutList[index].getSourceUrl());
                } catch (Exception e)
                {
                    this.logger.debug("unable to clean URL \""
                                + stageOutList[index].getSourceUrl() +
                                "\"",
                                e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageOut",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }
                try {
                    destinationUrl
                        = cleanUrl(stageOutList[index].getDestinationUrl());
                } catch (Exception e)
                {
                    this.logger.debug("unable to clean URL \""
                                + stageOutList[index].getDestinationUrl()
                                + "\"",
                                e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageOut",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }

                //sourceUrl
                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileStageOut sourceUrl before:"
                                +sourceUrl);
                }
                sourceUrl = resolveVariableInString(
                    "fileStageOut", sourceUrl);
                try {
                    GlobusURL url = new GlobusURL(sourceUrl);
                    if (url.getProtocol().equals("file")) {
                        sourceUrl = fsMap.mapFileName(url);
                    }
                } catch (java.net.MalformedURLException mue) {
                    sourceUrl = null;
                }

                if (sourceUrl == null) {
                    throw FaultUtils.createStagingFault(
                            "null",
                            destinationUrl,
                            "fileStageOut",
                            getState(),
                            (String) null,
                            (ResourceKey) getID());
                }

                try {
                    stageOutList[index].setSourceUrl(cleanUrl(sourceUrl));
                } catch (Exception e)
                {
                    this.logger.debug(
                        "unable to clean URL \"" + sourceUrl + "\"", e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageOut",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }

                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileStageOut sourceUrl after:"
                                +sourceUrl);
                }
                //END sourceUrl

                //destinationUrl
                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileStageOut destinationUrl before:"
                                +destinationUrl);
                }
                destinationUrl = resolveVariableInString(
                    "fileStageOut", destinationUrl);
                try {
                    GlobusURL url = new GlobusURL(destinationUrl);
                    if (url.getProtocol().equals("file")) {
                        destinationUrl = null;
                    }
                } catch (java.net.MalformedURLException mue) {
                    destinationUrl = null;
                }

                if (destinationUrl == null) {
                    throw FaultUtils.createStagingFault(
                            sourceUrl,
                            "null",
                            "fileStageOut",
                            getState(),
                            (String) null,
                            (ResourceKey) getID());
                }

                try {
                    stageOutList[index].setDestinationUrl(
                        cleanUrl(destinationUrl));
                } catch (Exception e)
                {
                    this.logger.debug(
                        "unable to clean URL \"" + destinationUrl +"\"", e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileStageOut",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }

                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileStageOut destinationUrl after:"
                                +destinationUrl);
                }
                //END destinationUrl
            }
        }
        //END fileStageOut

        //fileCleanUp
        DeleteRequestType deleteRequest = jobDescription.getFileCleanUp();
        if (deleteRequest != null)
        {
            if (fsMap == null)
            {
                fsMap = getFileSystemMapping();
            }

            DeleteType[] deletionList = deleteRequest.getDeletion();

            for (int index=0; index<deletionList.length; index++)
            {

                String fileUrl = null;
                try {
                    fileUrl = cleanUrl(deletionList[index].getFile());
                } catch (Exception e)
                {
                    this.logger.debug("unable to clean URL \""
                                + deletionList[index].getFile()
                                + "\"",
                                e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileCleanUp",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }
                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileCleanUp file before:"
                                +fileUrl);
                }
                fileUrl = resolveVariableInString(
                    "fileCleanUp", fileUrl);
                try {
                    GlobusURL url = new GlobusURL(fileUrl);
                    if (url.getProtocol().equals("file")) {
                        fileUrl = fsMap.mapFileName(url);
                    }
                } catch (java.net.MalformedURLException mue) {
                    this.logger.warn(mue);
                    fileUrl = null;
                }

                if (fileUrl == null) {
                    throw FaultUtils.createServiceLevelAgreementFault(
                            "fileCleanUp",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID());
                }

                try {
                    deletionList[index].setFile(cleanUrl(fileUrl));
                } catch (Exception e)
                {
                    this.logger.debug(
                        "unable to clean URL \"" + fileUrl + "\"", e);
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            "fileCleanUp",
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            e);

                    throw fault;
                }
                if (this.logger.isDebugEnabled()) {
                    this.logger.trace("fileCleanUp file after:"
                                +fileUrl);
                }
            }
        }
        //END fileStageOut

        str = jobDescription.getExecutable();
        str = resolveVariableInString("executable", str);
        str = str.replaceAll("/+", "/"); //normalize the path
        jobDescription.setExecutable(str);

        str = jobDescription.getDirectory();
        str = resolveVariableInString("directory", str);
        if (!str.startsWith("/") && !str.matches("^[a-zA-Z]:.*"))
        {
            str = "${GLOBUS_USER_HOME}" + File.separator + str;
            str = resolveVariableInString("directory", str);
        }
        str = str.replaceAll("/+", "/"); //normalize the path
        jobDescription.setDirectory(str);

        arr = jobDescription.getArgument();

        if (arr != null) {
            for (int i = 0; i < arr.length; i++) {
                arr[i] = resolveVariableInString("argument", arr[i]);
            }
        }
        jobDescription.setArgument(arr);

        env = jobDescription.getEnvironment();
        if (env != null) {
            for (int i = 0; i < env.length; i++) {
                env[i].setName(resolveVariableInString(
                       "environment",
                       env[i].getName()));
                env[i].setValue(resolveVariableInString(
                       "environment",
                       env[i].getValue()));
            }
        }
        jobDescription.setEnvironment(env);

        str = jobDescription.getStdin();
        str = resolveVariableInString("stdin", str);
        jobDescription.setStdin(str);

        str = jobDescription.getStdout();
        if (str == null) {
            this.logger.trace("No stdout set. Setting to /dev/null.");
            jobDescription.setStdout("/dev/null");
        } else {
            str = resolveVariableInString("stdout", str);

            //take care of relative path here since we need to set an RP with it
            if (!str.startsWith("/") && !str.matches("^[a-zA-Z]:.*"))
            {
                str = jobDescription.getDirectory() + File.separator + str;
                str = resolveVariableInString("stdout", str);
            }

            str = str.replaceAll("/+", "/");

            jobDescription.setStdout(str);
        }

        str = jobDescription.getStderr();
        if (str == null) {
            this.logger.trace("No stderr set. Setting to /dev/null.");
            jobDescription.setStderr("/dev/null");
        } else {
            str = resolveVariableInString("stderr", str);

            //take care of relative path here since we need to set an RP with it
            if (!str.startsWith("/") && !str.matches("^[a-zA-Z]:.*"))
            {
                str = jobDescription.getDirectory() + File.separator + str;
                str = resolveVariableInString("stderr", str);
            }

            str = str.replaceAll("/+", "/");

            jobDescription.setStderr(str);
        }

        arr = jobDescription.getLibraryPath();
        if (arr != null) {
            for (int i = 0; i < arr.length; i++) {
                arr[i] = resolveVariableInString("libraryPath", arr[i]);
            }
        }
        jobDescription.setLibraryPath(arr);

        //recreate the perl attributes with resolved variable values
        //initPerlJobDescription();

        this.logger.debug("Leaving resolveVariables()");
    }

    private String resolveVariableInString(String attribute, String s)
        throws                              ServiceLevelAgreementFaultType
    {
        if (s == null) {
            return null;
        }

        int index = 0;
        int nestingDepth = 0;
        int nestingIndex = 0;
        int                             dollar_idx;
        int                             close_idx = -1;
        StringBuffer                    buffer = new StringBuffer(s);
        String variableName = null;
        Object variableValue = null;

        this.logger.trace("resolving variables in attribute " + attribute);

        while (   (index < buffer.length())
               && ((dollar_idx = buffer.indexOf("$", index)) != -1))
        {
            this.logger.trace("looking at string " + buffer.toString());
            this.logger.trace("found $ at index "
                             + Integer.toString(dollar_idx));

            if (   (buffer.length() > dollar_idx+3) // at least '${?}'
                && (buffer.charAt(dollar_idx+1) == '{'))
            {
                this.logger.trace("found '{'---looks like a reference");

                if (index < nestingIndex)
                {
                    //re-checking the section we just swapped in for a variable
                    nestingDepth++;

                    //abort if we've reached the maximum nesting limit
                    if (nestingDepth >= MAX_VARIABLE_DEPTH)
                    {
                        Exception cause = new Exception(
                            i18n.getMessage(
                                Resources.VARIABLE_DEPTH_EXCEEDED,
                                new String[] {
                                    variableName,
                                    variableValue.toString() }));
                        ServiceLevelAgreementFaultType fault
                            = FaultUtils.createServiceLevelAgreementFault(
                                attribute,
                                getState(),
                                (String) null,
                                0,
                                (ResourceKey) getID(),
                                cause);

                        throw fault;
                    }
                }
                else
                {
                    nestingDepth = 0;
                }

                close_idx = buffer.indexOf("}");

                if (close_idx == -1)
                {
                    this.logger.trace("didn't find '}'---bailing");

                    return buffer.toString();
                }

                variableName = buffer.substring(
                        dollar_idx+2,
                        close_idx);

                this.logger.trace("looking up " + variableName
                            + " in " + variableMap);

                variableValue = variableMap.get(variableName);

                this.logger.trace("mapped " + variableName + " to value "
                                 +variableValue);

                if (variableValue == null)
                {
                    Exception cause = new Exception(
                        i18n.getMessage(
                            Resources.NULL_VARIABLE_VALUE,
                            new String[] { variableName }));
                    ServiceLevelAgreementFaultType fault
                        = FaultUtils.createServiceLevelAgreementFault(
                            attribute,
                            getState(),
                            (String) null,
                            0,
                            (ResourceKey) getID(),
                            cause);

                    throw fault;
                }

                buffer.replace(
                    dollar_idx,
                    close_idx+1,
                    variableValue.toString());

                //beginning of substring we just swapped in
                index = dollar_idx;

                //end of substring we just swapped in
                nestingIndex = index + variableValue.toString().length();
            } else {
                this.logger.trace("No '{' after $, not a variable");
                index++;
            }
        }

        s = buffer.toString();

        this.logger.trace("Final string is " + s);

        return s;
    }

    public synchronized void start()
    {
        boolean started
            = ((Boolean)getResourceDatum(ManagedJobConstants.RD_STARTED))
            .booleanValue();
        if (!started)
        {
            if (this.logger.isDebugEnabled()) {
                this.logger.trace("Adding to RunQueue.");
            }
            setResourceDatum(ManagedJobConstants.RD_STARTED, new Boolean(true));
        }
        RunQueue.getInstance().add((ResourceKey)this.getID());
    }

    FileMapping getFileSystemMapping()
        throws                              FileMappingException
    {
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String globusLocation = factoryHomeConfiguration.getGlobusLocation();

        ManagedExecutableJobResourceDataType resourceData
            = (ManagedExecutableJobResourceDataType) getResourceData();
        String localResourceManager =
            ((ManagedExecutableJobResourceDataType)resourceData)
                .getLocalResourceManager();

        String filePathConfig = new File(globusLocation
                + File.separator + "etc"
                + File.separator + "gram-service",
                FileMapping.FILE_MAP_CONFIG_FILE).toString();

        return new FileMapping(filePathConfig, localResourceManager);
    }

    public void setState(StateEnumeration state)
        throws Exception
    {
        if (this.logger.isDebugEnabled()) {
            this.logger.trace("receiving request for state change to " +
                         state.getValue());
        }

        //set the new state value
        super.setState(state);

        if (this.logger.isDebugEnabled()) {
            this.logger.debug("State of job " + this.getID()
                        + " changed to: " + state.toString());
        }

        //fetch the holding flag
        Boolean holding = (Boolean) getResourcePropertySet().get(
            ManagedJobConstants.RP_HOLDING).get(0);
        if (this.logger.isDebugEnabled()) {
            this.logger.trace("Holding: " + holding);
        }

        //fetch the fault if the state is 'Failed'
        ResourceProperty faultRP
            = this.getResourcePropertySet().get(ManagedJobConstants.RP_FAULT);
        FaultResourcePropertyType fault = null;
        if (state.equals(StateEnumeration.Failed))
        {
            fault = (FaultResourcePropertyType) faultRP.get(0);
        }

        //fetch the exit code if the state is Done or Failed
        ResourceProperty exitCodeRP = getResourcePropertySet().get(
            ManagedExecutableJobConstants.RP_EXIT_CODE);
        Integer exitCode = null;
        if (state.equals(StateEnumeration.Done)
                || state.equals(StateEnumeration.Failed))
        {
            exitCode = (Integer) exitCodeRP.get(0);
        }

        //create notification message
        StateChangeNotificationMessageType message
            = new StateChangeNotificationMessageType();
        message.setState(state);
        message.setHolding(holding.booleanValue());
        message.setFault(fault);

        if (exitCode != null) {
            this.logger.trace("Setting exitCode to " + exitCode.toString());
            message.setExitCode(exitCode.intValue());
        } else {
            this.logger.trace("exitCode is null");
        }

        //save the notification for sending to late subscribers

        StateChangeNotificationMessageWrapperType messageWrapper
            = new StateChangeNotificationMessageWrapperType();
        messageWrapper.setStateChangeNotificationMessage(message);

        //notify on the topic using the state change message
        LinkedList topicPath = new LinkedList();
        topicPath.add(ManagedJobConstants.RP_STATE);
        Topic stateChangeTopic = this.getTopicList().getTopic(topicPath);
        this.logger.trace("Notifying of job state change to topic listeners");

        stateChangeTopic.notify(messageWrapper);
    }

    public JobStateMonitor getJobStateMonitor() {
        if (this.jobMonitor == null) {
            this.logger.trace("job monitor is null");
            try {
                ManagedJobFactoryHome factoryHome
                    = ManagedJobFactoryHome.getHome();

                String localResourceManager
                    = ((ManagedExecutableJobResourceDataType)
                    getResourceData()).getLocalResourceManager();
                ManagedJobFactoryResource factoryResource =
                    (ManagedJobFactoryResource) factoryHome.find(
                        ManagedJobHelper.getResourceKey(localResourceManager));
                this.jobMonitor = (JobStateMonitor)
                        factoryResource.getJobStateMonitor();
            } catch (NamingException ne) {
                this.logger.warn(ne);
            } catch (ResourceException re) {
                this.logger.warn(re);
            }
        }
        return this.jobMonitor;
    }

    public void setScriptEnvironment(String[] env) {
        this.scriptEnvironment = env;
    }

    public String[] getScriptEnvironment() {
        return this.scriptEnvironment;
    }

    public void setScript(JobManagerScript script) {
        this.script = script;
    }

    public JobManagerScript getScript() {
        return this.script;
    }

    public void setPerlJobDescription(PerlJobDescription perlJobDescription) {
        this.perlJobDescription = perlJobDescription;
    }

    public PerlJobDescription getPerlJobDescription() {
        return this.perlJobDescription;
    }

    public EndpointReferenceType getReliableFileTransferFactoryEndpoint()
    {
        if (this.reliableFileTransferFactoryEndpoint == null)
        {
            //get RFT factory
            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String name = "ReliableFileTransferFactoryService";
            String nameOverride
                = factoryHomeConfiguration.getStagingFactoryName();
            if (nameOverride != null)
            {
                name = nameOverride;
            }

            String localResourceManager = (String) getResourceDatum(
                ManagedExecutableJobConstants.RD_LOCAL_RESOURCE_MANAGER);
            SimpleResourceKey factoryKey
                = ManagedJobHelper.getResourceKey(localResourceManager);
            ManagedJobFactoryResource factoryResource = null;
            try
            {
                factoryResource
                    = (ManagedJobFactoryResource) ManagedJobFactoryHome
                    .getHome().find(factoryKey);
            } catch (Exception e)
            {
                throw new RuntimeException(e);
            }
            EndpointReferenceType endpoint
                = (EndpointReferenceType) factoryResource
                .getStagingDelegationFactoryEndpoint();
            AttributedURI address = endpoint.getAddress();
            File path = new File(address.getPath());
            URL serviceUrl = null;
            String wsTransportProtocol
                = factoryHomeConfiguration.getStagingProtocol();
            try
            {
               String scheme = null;
                if (wsTransportProtocol == null)
                {
                    scheme = "local";
                }
                else
                {
                    scheme = address.getScheme();
                }
                serviceUrl = new URL(
                    scheme,
                    address.getHost(),
                    address.getPort(),
                    path.getParent() + "/" + name);

                if (logger.isDebugEnabled())
                {
                    logger.debug("RFT Factory EPR: " + serviceUrl.toString());
                }
            } catch (Exception e)
            {
                String errorMessage =
                    i18n.getMessage(Resources.STAGING_EPR_CREATION_ERROR,
                                    serviceUrl);
                throw new RuntimeException(errorMessage, e);
            }

            try {
                this.reliableFileTransferFactoryEndpoint
                    = AddressingUtils.createEndpointReference(
                        serviceUrl.toString(), null);
            } catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.STAGING_EPR_CREATION_ERROR,
                                    serviceUrl);
                throw new RuntimeException(errorMessage, e);
            }
        }

        return this.reliableFileTransferFactoryEndpoint;
    }

    private EndpointReferenceType reliableFileTransferFactoryEndpoint = null;

    public Object getStateTransitionMutex()
    {
        return stateTransitionMutex;
    }
    private Object stateTransitionMutex = new Object();

    //========= Security Descriptor =========

    private ResourceSecurityDescriptor securityDescriptor = null;

    public ResourceSecurityDescriptor getSecurityDescriptor() {
        if (this.securityDescriptor != null) {
            this.logger.trace("MJR security descriptor authz: "
                        + this.securityDescriptor.getAuthz());
        } else {
            this.logger.trace("MRJR security descriptor is null");
        }
        return this.securityDescriptor;
    }

    public Log getLogger()
    {
        return this.logger;
    }

    //==== resource data getters/setters ====
    public InternalStateEnumeration getInternalState()
    {
       if (logger.isDebugEnabled())
       {
            this.logger.debug("getting resource datum internalState");
        }

        InternalStateEnumeration internalState = null;

        ManagedExecutableJobResourceDataType resourceData
            = (ManagedExecutableJobResourceDataType) getResourceData();

        this.logger.trace("Obtaining lock on resourceData");
        synchronized (resourceData)
        {
            this.logger.trace("Obtained lock on resourceData");

            internalState = resourceData.getInternalState();
            this.logger.trace("Releasing lock on resourceData");
        }
        this.logger.trace("Released lock on resourceData");

        return internalState;
    }

    public void setInternalState(
        InternalStateEnumeration            internalState)
        throws                              ResourceException
    {
        Log logger = getLogger();

        if (logger.isDebugEnabled())
        {
            logger.debug(
                "Setting resource datum internalState to value "
                + internalState);
        }

        ManagedExecutableJobResourceDataType resourceData
            = (ManagedExecutableJobResourceDataType) getResourceData();
        logger.trace("Obtaining lock on resourceData");
        synchronized (resourceData)
        {
            logger.trace("Obtained lock on resourceData");

            resourceData.setInternalState(internalState);

            logger.trace("Releasing lock on resourceData");
        }
        logger.trace("Released lock on resourceData");
    }

}
