/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;


import java.rmi.RemoteException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.oasis.wsrf.properties.ResourceUnknownFaultType;

import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.ReleaseInputType;
import org.globus.exec.generated.ReleaseOutputType;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.util.I18n;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.ResourceKey;

public class ManagedExecutableJobService
{
    private static Log logger =
        LogFactory.getLog(ManagedExecutableJobService.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    public ReleaseOutputType release(
        ReleaseInputType                    parameters)
        throws                              RemoteException,
                                            FaultType
    {
        ManagedExecutableJobResource resource = getResource();
        ResourceKey resourceKey = (ResourceKey) resource.getID();

        Log localLogger = LogFactory.getLog(
            ManagedExecutableJobService.class.getName()
            + "." + resourceKey.getValue().toString());

        if (localLogger.isTraceEnabled())
        {
            localLogger.debug("Entering release()");
        }

        StateMachine.release(getResource());

        if (localLogger.isTraceEnabled())
        {
            localLogger.debug("Leaving release()");
        }

        return new ReleaseOutputType();
    }

    private static ManagedExecutableJobResource getResource()
        throws                              RemoteException
    {
        ResourceHome home = null;
        ResourceKey key = null;
        ManagedExecutableJobResource resource = null;
        try {
            ResourceContext ctx = ResourceContext.getResourceContext();
            home = ctx.getResourceHome();
            key = ctx.getResourceKey();
            resource = (ManagedExecutableJobResource) home.find(key);
        } catch (Exception e) {
            String errorMessage =
                i18n.getMessage(Resources.RESOURCE_FIND_ERROR);
            logger.error(errorMessage, e);
            throw new RemoteException(errorMessage, e);
        }

        if (key == null || resource == null) {
            throw new ResourceUnknownFaultType();
        }

        return resource;
    }
}
