/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.List;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Date;

import javax.naming.NamingException;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.ManagedExecutableJobResourcePropertiesType;
import org.globus.exec.generated.ManagedExecutableJobResourceStateType;
import org.globus.exec.generated.ManagedExecutableJobResourceDataType;
import org.globus.exec.generated.TopicListenerDescriptionType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.FilePairType;
import org.globus.exec.generated.InternalStateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.NameValuePairType;
import org.globus.exec.generated.StateChangeNotificationMessageType;
import org.globus.exec.generated.StateChangeNotificationMessageWrapperType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.monitoring.JobStateMonitor;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.service.job.PersistentManagedJobResource;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.factory.ManagedJobFactoryResource;
import org.globus.exec.service.utils.SubscriptionPersistenceUtils;
import org.globus.exec.service.utils.PersistenceHelper;

import org.globus.exec.utils.Resources;

import org.globus.util.I18n;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.PersistentResource;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.impl.PersistentReflectionResource;
import org.globus.wsrf.impl.DirtyFlagHolder;
import org.globus.wsrf.impl.PersistentResourceProperty;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.XmlPersistenceHelper;
import org.globus.wsrf.utils.XmlUtils;

import org.ietf.jgss.GSSException;

import org.oasis.wsrf.faults.BaseFaultType;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class PersistentManagedExecutableJobResource
    extends     ManagedExecutableJobResource
    implements  PersistentManagedJobResource
{
    private static Log logger = LogFactory.getLog(
        PersistentManagedExecutableJobResource.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    /**
     * The dirty flag is set to true at instanciation time, so that the
     * initial creator of the resource just needs to call {@link #store() store}
     * in order to actually persist the new resource for the first time.
     */
    private boolean dirty = true; //in most cases,
    //creation will occur after a persisted version of the resource exists
    //already in the persistence store.

    private Object dirtyCond = new Object();

    public void setResourceProperty(
        QName                               name,
        Object                              value)
    {
        synchronized (dirtyCond)
        {
            super.setResourceProperty(name, value);
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, name);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    public void setResourceDatum(
        String                              name,
        Object                              value)
    {
        synchronized (dirtyCond)
        {
            super.setResourceDatum(name, value);
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, name);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    public void initializeWholeState(
        Object                                  wholeResourceState,
        QName                                   resourceElementQName,
        Object                                  key)
        throws                                  ResourceException
    {
        //this is just to reload the internal rendezvous state data.
        //We do not need/want to do this is the resource is not persistent.

        super.initializeWholeState(
            wholeResourceState, resourceElementQName, key);

        ManagedExecutableJobResourceStateType wholeJobState =
            (ManagedExecutableJobResourceStateType)wholeResourceState;

        this.setInternalRendezvousState(wholeJobState.getResourceData().
                                        getInternalRendezvousData());
    }

    public void setFault(FaultType fault)
    {
        synchronized (dirtyCond)
        {
            super.setFault(fault);
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, fault);
                throw new RuntimeException(errorMessage, e);
            }
        }
    }

    public void setState(StateEnumeration state)
        throws Exception
    {
        synchronized (dirtyCond)
        {
            super.setState(state);

            //here we probably want to persist last state timeStamp to disk too,
            //and not necessarily via an RP.
            //persist the resource properties to disk

            //Alain claimed that this is "not needed since RP State does it via
            //PersistentRPs", but without it Done can be skipped when high loads
            //are put on the service.  Not sure what's going on.
            setDirty(true);

            this.store();
        }
    }

    public Class getWholeStateType() {
        return ManagedExecutableJobResourceStateType.class;
    }

    public QName getWholeStateQName() {
        return WHOLE_STATE_ELEMENT;
    }

    private static final QName WHOLE_STATE_ELEMENT = new QName(
                        ManagedExecutableJobConstants.EXEC_JOB_STATE_NS,
                        "ManagedExecutableJobResourceState");

    public QName getResourcePropertiesQName() {
        return ManagedJobConstants.RESOURCE_PROPERTY_SET;
    }

    /**
     * Called by PersistenceHelper when store() is called
     * @return Object the whole serializable state of the resource
     */
    public Object getWholeState() {
        ManagedExecutableJobResourceStateType resourceState =
            new ManagedExecutableJobResourceStateType();
        //resource properties state
        resourceState.setResourceProperties(
            (ManagedExecutableJobResourcePropertiesType)this.getResourceBean());
        //internal job implementation state
        resourceState.setResourceData(
            (ManagedExecutableJobResourceDataType) getResourceData());

        //add internal rendezvous implementation state before persisting.
        //we could put this in an overriden getResourceData(), but it's not
        //needed for every call (well, overhead might be small anyway...).
        //store() might be good spot to put it, but I'd like to keep store()
        //simple...
        getResourceData().setInternalRendezvousData(
            this.getInternalRendezvousState());

        return resourceState;
    }

    /**
     * <p>
     * If the resource was successfully stored, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     * @throws ResourceException if the resource could not be stored
     */
    public void store() throws ResourceException
    {
        getPersistenceHelper().store();
    }

    /**
     * <p>
     * If the resource was successfully loaded, its dirty flag is set to false.
     * <p>
     * <b>Postcondition</b> getDirty() == false
     * @see #setDirty(boolean) setDirty()
     *
     */
    public void load(ResourceKey key) throws ResourceException {
        getPersistenceHelper().load(key);
    }

    public void remove() throws ResourceException {
        super.remove();
        getPersistenceHelper().remove();
    }

    private PersistenceHelper getPersistenceHelper()
        throws ResourceException{
        if (this.persistenceHelper == null) {
            try {
                this.persistenceHelper = new PersistenceHelper(this);
            }
            catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.PERSISTENCE_HELPER_CREATION_ERROR);
                logger.error(errorMessage, e);
                throw new ResourceException(errorMessage, e);
            }
        }
        return this.persistenceHelper;
    }

    private PersistenceHelper persistenceHelper;

    protected ResourceProperty createNewResourceProperty(
        QName rpQName,
        Object resourceBean)
        throws Exception {
        ResourceProperty prop =
            super.createNewResourceProperty(rpQName, resourceBean);

        return new PersistentResourceProperty(prop, this);
    }

    /**
     * Sets the dirty flag on this persistent object. The resource will not be
     * persisted unless it the firty flag is true.
     *
     * @param changed boolean To flag the resource as changed i.e. "dirty" and
     * have it persisted by the next invocation of {@link #store() store}
     */
    public void setDirty(boolean changed) {
        if (logger.isDebugEnabled()) {
            logger.debug("setting dirty flag to " + changed);
        }
        this.dirty = changed;
    }

    public boolean getDirty() {
        return this.dirty;
    }

    public void setInternalState(
        InternalStateEnumeration            internalState)
        throws                              ResourceException
    {
        synchronized (dirtyCond)
        {
            super.setInternalState(internalState);
            setDirty(true);
            try {
                store();
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.POST_UPDATE_PERSISTENCE_ERROR, "internalState");
                throw new RuntimeException(errorMessage, e);
            }
        }
    }
}
