/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.exec.utils.Resources;

import org.globus.util.I18n;
import org.globus.wsrf.ResourceKey;

class RunThread
    extends                                 Thread
{
    /** Log4J logger */
    private static Log logger = LogFactory.getLog(RunThread.class);
    static I18n i18n = I18n.getI18n(Resources.class.getName());

    private boolean done = false;

    private RunQueue runQueue;

    RunThread(String           name, 
              RunQueue         queue)
    {
        super(name);
        this.runQueue = queue;
    }

    public synchronized void terminate()
    {
        this.done = true;
    }

    public void run()
    {
        while (!this.done)
        {
            ResourceKey resourceKey;

            synchronized (this.runQueue)
            {
                while ((!this.done) && (this.runQueue.size() == 0))
                {
                    try
                    {
                        this.runQueue.wait();
                    }
                    catch (InterruptedException e) { }
                }
            }

            do
            {
                synchronized (this.runQueue)
                {
                    if (this.runQueue.size() > 0)
                    {
                        resourceKey = (ResourceKey) this.runQueue.remove();
                    }
                    else
                    {
                        resourceKey = null;
                    }
                }

                if (resourceKey != null)
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Processing state for "
                            + resourceKey.toString());
                    }

                    try
                    {
                        StateMachine.processState(resourceKey);
                    }
                    catch (Throwable e)
                    {
                        logger.error(i18n.getMessage(
                            Resources.STATE_TRANSITION_FAILED), e);
                    }

                    if (logger.isDebugEnabled())
                    {
                        logger.debug("processState returned");
                    }

                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Done processing state for "
                                + resourceKey.toString());
                    }
                }
            }
            while ((!done) && (resourceKey != null));
        }
    }
}
