/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.TimerTask;
import java.util.Vector;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.addressing.ReferencePropertiesType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;
import org.globus.exec.utils.Resources;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.core.notification
       .ResourcePropertyValueChangeNotificationElementType;
import org.globus.wsrf.encoding.ObjectDeserializer;

import org.oasis.wsrf.properties.ResourcePropertyValueChangeNotificationType;

/**
 * This thread class is needed for use be StateMachine to avoid a race
 * condition between the JobStateMonitor.dispatchEvent() and
 * JobStateMonitor.unregisterJobID() methods.
 */
public class StagingListener
    implements                 NotifyCallback, Serializable
{
    /** Log4J logger */
    static Log logger = LogFactory.getLog(StagingListener.class);

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static StagingListener INSTANCE;
    protected static Hashtable execJobMap;
    private static Hashtable statusMap;
    //protected static Timer pollTimer;
    static {
        INSTANCE = new StagingListener();
        execJobMap = new Hashtable();
        statusMap = new Hashtable();
        //pollTimer = null;
    }

    private StagingListener() { }

    public static StagingListener getInstance()
    {
        return INSTANCE;
    }

    public static synchronized StagingListener registerTransferJob(
        EndpointReferenceType               transferJobEndpoint,
        ResourceKey                         execJobKey)
    {
        String transferJobKey = endpointToKey(transferJobEndpoint);

        if (logger.isDebugEnabled())
        {
            logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                "Entering registerTransferJob()");
        }

        //start polling every 5 minutes
        /*
        if (pollTimer == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.debug("starting poll timer");
            }
            pollTimer = new Timer(true);
            pollTimer.schedule(new PollTask(), 300000, 300000);
        }
        */

        //add mapping from the request key to the resource key
        execJobMap.put(transferJobKey, execJobKey);

        logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
            "Leaving registerTransferJob()");

        return INSTANCE;
    }

    public static synchronized void unregisterTransferJob(
        EndpointReferenceType               transferJobEndpoint)
    {
        String transferJobKey = endpointToKey(transferJobEndpoint);

        logger.debug(getLoggingPrefix(null, transferJobKey) +
            "Entering unregisterTransferJob()");

        synchronized (execJobMap)
        {
            execJobMap.remove(transferJobKey);

            //cancel timer if there are no more transfer jobs
            /*
            if (execJobMap.size() == 0)
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("stopping poll timer");
                }
                pollTimer.cancel();
                pollTimer = null;
            }
            */
        }

        synchronized (statusMap)
        {
            statusMap.remove(transferJobKey);
        }

        logger.debug(getLoggingPrefix(null, transferJobKey) +
            "Leaving unregisterTransferJob()");
    }

    public static synchronized RequestStatusType getNextStatus(
        EndpointReferenceType               transferJobEndpoint)
    {
        String transferJobKey = endpointToKey(transferJobEndpoint);
        RequestStatusType status = null;
        synchronized (statusMap)
        {
            Vector statusEntries = (Vector) statusMap.get(transferJobKey);
            if (statusEntries == null)
            {
                return null;
            }

            status = (RequestStatusType) statusEntries.remove(0);
            if (statusEntries.size() == 0)
            {
                statusMap.remove(transferJobKey);
            }
        }

        return status;
    }

    public synchronized void deliver(
        List                                topicPath,
        EndpointReferenceType               producerEndpoint,
        Object                              messageWrapper)
    {
        String transferJobKey = endpointToKey(producerEndpoint);
        ResourceKey execJobKey = (ResourceKey) execJobMap.get(transferJobKey);

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                "Entering deliver()");
        }

        if (execJobKey == null)
        {
            logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                "No resource mapping for transfer job.");
            return;
        }

        if (logger.isDebugEnabled())
        {
            logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                "Received notification from RFT.");
        }


        try
        {
            ResourcePropertyValueChangeNotificationType message
                = ((ResourcePropertyValueChangeNotificationElementType)
                messageWrapper).getResourcePropertyValueChangeNotification();

            RequestStatusType status = null;
            try {
              status = (RequestStatusType)
                 ObjectDeserializer.getFirstAsObject(
                                 message.getNewValue(), RequestStatusType.class);
            } catch (Exception e) {
                try {
                    status = (RequestStatusType)message.getNewValue().get_any()[0]
                                .getObjectValue(RequestStatusType.class);
                } catch (Exception ex) {
                    logger.error(
                    	"Notification from RFT can't be deserialized properly: "+
                    	ex.toString());
                }
            }

            // check if notification must be processed
            RequestStatusTypeEnumeration requestState =
                     status.getRequestStatus();
            if (requestState != null)
            {
                if (  !(requestState.equals(RequestStatusTypeEnumeration.Done)
                    ||requestState.equals(RequestStatusTypeEnumeration.Failed)))
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug(
                            getLoggingPrefix(execJobKey, transferJobKey) +
                             "Notification is irrelevant ("
                             + requestState.getValue()
                             + ") and will not be processed");
                    }
                    return;
                }
                else
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug(
                            getLoggingPrefix(execJobKey, transferJobKey) +
                            "Notification is relevant ("
                            + requestState.getValue() + ")");
                    }
                }
            }
                
            synchronized (statusMap)
            {
                Vector statusEntries  = (Vector) statusMap.get(transferJobKey);
                if ((statusEntries == null))
                {
                    statusEntries = new Vector();
                    statusMap.put(transferJobKey, statusEntries);
                }

                statusEntries.add(status);
            }

            RunQueue.getInstance().add(execJobKey);

            if (logger.isDebugEnabled())
            {
                logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                    "Processed notification. Job is in RunQueue.");
            }
        }
        catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                            Resources.STATE_NOTIFICATION_PROCESSING_ERROR);
            logger.error(errorMessage, e);
        }

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(execJobKey, transferJobKey) +
                "Leaving deliver()");
        }
    }

    private static String endpointToKey(
        EndpointReferenceType               transferJobEndpoint)
    {
        String transferJobKey = null;

        try {
            ReferencePropertiesType referenceProperties
                = transferJobEndpoint.getProperties();
            MessageElement[] any = referenceProperties.get_any();
            //transferJobKey = AnyHelper.toString(any[0];
            transferJobKey = any[0].getValue();
        } catch (Exception e)
        {
            //TODO internationalize
            throw new RuntimeException("notification missing RFT resource ID");
        }

        return transferJobKey;
    }

    static String getLoggingPrefix(
        ResourceKey                         execJobKey,
        String                              transferJobKey)
    {
        if (execJobKey != null)
        {
            return new String(
                  "[execJobKey:" + execJobKey.getValue() + ","
                + "transferJobKey:" + transferJobKey + "] ");
        }

        return new String();
    }
}

class PollTask extends TimerTask
{
    public void run()
    {
        synchronized (StagingListener.execJobMap)
        {
            Iterator transferJobKeyIterator
                = StagingListener.execJobMap.keySet().iterator();
            while (transferJobKeyIterator.hasNext())
            {
                String transferJobKey = (String) transferJobKeyIterator.next();
                ResourceKey execJobKey = (ResourceKey)
                    StagingListener.execJobMap.get(transferJobKey);

                if (StagingListener.logger.isDebugEnabled())
                {
                    StagingListener.logger.debug(
                        StagingListener.getLoggingPrefix(
                            execJobKey,
                            transferJobKey) +
                        "polling for transfer job status...");
                }

                /* StateMachine is setup to poll for status if it enters
                   one of the staging response transition methods and the
                   StagingListener.getNextStatus() method returns null.
                */
                RunQueue.getInstance().add(execJobKey);
            }
        }
    }
}
