/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.rmi.RemoteException;
import java.util.Calendar;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.exec.generated.FaultType;
import org.globus.exec.service.exec.ManagedExecutableJobResource;
import org.globus.exec.service.exec.StagingListener;
import org.globus.exec.service.utils.DelegatedCredential;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.utils.Resources;
import org.globus.rft.generated.BaseRequestType;
import org.globus.rft.generated.CreateReliableFileTransferInputType;
import org.globus.rft.generated.CreateReliableFileTransferOutputType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.Start;
import org.globus.rft.generated.StartOutputType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.util.I18n;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.globus.exec.utils.service.ManagedJobHelper;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.service.exec.utils.LocalInvocationHelper;

public class LocalStagingHelper extends StagingHelper
{
    private static Log logger =
        LogFactory.getLog(LocalStagingHelper.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
        
    public EndpointReferenceType submitStagingRequest(
        ManagedExecutableJobResource        resource,
        BaseRequestType                     stagingRequest)
        throws                              Exception
    {
        logger.trace(getLoggingPrefix(resource) +
            "Entering sumbitStagingRequest()");

        checkForCredential(resource);

        //setup the create input
        CreateReliableFileTransferInputType createInput
            = new CreateReliableFileTransferInputType();
        if (stagingRequest instanceof TransferRequestType) {
            createInput.setTransferRequest((TransferRequestType)stagingRequest);
        } else {
            createInput.setDeleteRequest((DeleteRequestType)stagingRequest);
        }
        Calendar terminationTime = Calendar.getInstance();
        terminationTime.add(Calendar.HOUR, 12);
        createInput.setInitialTerminationTime(terminationTime);

        logger.debug("creating RFT resource");

        //create the transfer
        CreateReliableFileTransferOutputType createOutput = 
          (CreateReliableFileTransferOutputType) LocalInvocationHelper
              .getInstance()
                .callService(
                  "ReliableFileTransferFactoryService",
                  "createReliableFileTransfer",
                  (String) resource
                            .getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT),
                  new Object[] {createInput},
                  null);
        
        EndpointReferenceType transferJobEndpoint
            = createOutput.getReliableTransferEPR();
       
        logger.debug("created RFT resource");
        
        // start the transfer and setup the notifications
        StartOutputType startResponse
            = startTransferJob(resource, transferJobEndpoint);

        logger.debug(getLoggingPrefix(resource) +
            "Leaving sumbitStagingRequest()");

        return transferJobEndpoint;
    }

    public StartOutputType startTransferJob(
        ManagedExecutableJobResource            resource,
        EndpointReferenceType                   transferJobEndpoint)
        throws                                  Exception
    {
        logger.trace(getLoggingPrefix(resource) +
            "Entering startTransferJob()");

        checkForCredential(resource);

        //register with the staging listener
        StagingListener.registerTransferJob(
            transferJobEndpoint, (ResourceKey) resource.getID());

        Subscribe subscriptionRequest = new Subscribe();
        TopicExpressionType topicExpression = new TopicExpressionType(
            WSNConstants.SIMPLE_TOPIC_DIALECT,
            RFTConstants.REQUEST_STATUS_RESOURCE);
        subscriptionRequest.setTopicExpression(topicExpression);
        
        logger.debug("subscribing for state changes of RFT resource");

        // subscribe
        EndpointReferenceType subscriptionEndpoint
            = LocalInvocationHelper.getInstance().subscribe(
                   "ReliableFileTransferService",
                   ManagedJobHelper.getResourceKey(transferJobEndpoint),
                   (String) resource.getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT),
                   subscriptionRequest,
                   (NotifyCallback) StagingListener.getInstance());
                      
        logger.debug("subscribed for state changes of RFT resource");
        
        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT,
            subscriptionEndpoint);

        logger.debug("starting transfer");
        //start transfer
        StartOutputType startResponse = 
          (StartOutputType) LocalInvocationHelper.getInstance().callService(
              "ReliableFileTransferService",
              "start",
              (String) resource.getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT),
              new Object[] {new Start()},
              ManagedJobHelper.getResourceKey(transferJobEndpoint));
        
        logger.debug("started transfer");

        return startResponse;
    }

    public void destroyTransferJob(
        ManagedExecutableJobResource        resource,
        EndpointReferenceType               transferJobEndpoint)
        throws                              Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) + "Entering destroyTransferJob()");
        }

        checkForCredential(resource);

        //destroy status notification subscription
        EndpointReferenceType subscriptionEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT);
        
        if (subscriptionEndpoint != null)
        {
            resource.setResourceDatum(
               ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT,
               null);
            logger.debug("destroying subscription resource");
            // destroy subscription endpoint
            LocalInvocationHelper.getInstance().destroyResource(
                "SubscriptionManagerService",
                ManagedJobHelper.getResourceKey(subscriptionEndpoint),
                (String) resource.getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT));
            logger.debug("destroyed subscription resource");
        }

        if (transferJobEndpoint != null)
        {
            //unregister the transfer from the staging listener
            StagingListener.unregisterTransferJob(transferJobEndpoint);

            logger.debug("destroying rft resource");
            //destroy the transfer
            LocalInvocationHelper.getInstance().destroyResource(
                "ReliableFileTransferService",
                ManagedJobHelper.getResourceKey(transferJobEndpoint),
                (String) resource.getResourceProperty(ManagedJobConstants.RP_USER_SUBJECT));
            logger.debug("destroyed rft resource");
        }

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) + "Leaving destroyTransferJob()");
        }
    }


    public RequestStatusType poll(
        ManagedExecutableJobResource        resource,
        EndpointReferenceType               transferJobEndpoint)
    {
        GetResourcePropertyResponse response = null;
        RequestStatusType requestStatus = null;

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) + "Entering poll()");
        }

        try {
            checkForCredential(resource);

            if (transferJobEndpoint != null)
            {

                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resource) +
                        "polling for transfer job status.");
                }

                // polling for transfer status
                response =
                	LocalInvocationHelper.getInstance().getResourceProperty(
                         "ReliableFileTransferService",
                         ManagedJobHelper.getResourceKey(transferJobEndpoint),
                         RFTConstants.REQUEST_STATUS_RESOURCE, 
                         (String) resource.getResourceProperty(
                        		     ManagedJobConstants.RP_USER_SUBJECT));
                MessageElement[] any = response.get_any();
                try {
                    requestStatus =
                      (RequestStatusType) ObjectDeserializer.
                         toObject(any[0], RequestStatusType.class);
                } catch (Exception e) {
                    try {
                       requestStatus = (RequestStatusType)any[0]
                                          .getObjectValue(RequestStatusType.class);
                    } catch (Exception ex) {
                       logger.warn(
                    	   "Polling for requestStatus of transfer resource failed: " +
                            ex.toString());
                    } 
                }
            }
        } catch (Exception e) {
            logger.warn("unable to poll for transfer job status", e);
        }

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) + "Leaving poll()");
        }

        return requestStatus;
    }

    private void checkForCredential (
       ManagedExecutableJobResource resource)
       throws RemoteException  
    { 
        DelegatedCredential stagingCredential = null;
        try
        {
            stagingCredential = resource.getStagingCredential();
        }
         catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                Resources.CREDENTIAL_ACQUISITION_ERROR);
            FaultType fault = FaultUtils.makeFault(
                FaultType.class,
                (ResourceKey) resource.getID(),
                errorMessage,
                e,
                resource.getState(),
                null,
                0);
            resource.setFault(fault);
        }
        if (stagingCredential == null)
        {
            String errorMessage
                = i18n.getMessage(Resources.MISSING_STAGING_CREDENTIAL);
            throw new RemoteException(errorMessage);
        }
    }
}
 
