/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import javax.security.auth.Subject;
import javax.xml.rpc.Stub;
import javax.xml.soap.SOAPElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.exec.generated.FaultType;
import org.globus.exec.service.exec.ManagedExecutableJobHome;
import org.globus.exec.service.exec.ManagedExecutableJobResource;
import org.globus.exec.service.exec.StagingListener;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.utils.DelegatedCredential;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.rft.generated.BaseRequestType;
import org.globus.rft.generated.CreateReliableFileTransferInputType;
import org.globus.rft.generated.CreateReliableFileTransferOutputType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.ReliableFileTransferFactoryPortType;
import org.globus.rft.generated.ReliableFileTransferPortType;
import org.globus.rft.generated.Start;
import org.globus.rft.generated.StartOutputType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.service.
    ReliableFileTransferFactoryServiceAddressingLocator;
import org.globus.rft.generated.service.
    ReliableFileTransferServiceAddressingLocator;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.util.I18n;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.Authorization;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.GSISecureMsgAuthMethod;
import org.globus.wsrf.impl.security.descriptor.GSITransportAuthMethod;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.utils.PerformanceLog;
import org.ietf.jgss.GSSCredential;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsn.SubscriptionManager;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsn.WSBaseNotificationServiceAddressingLocator;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;

public class WSStagingHelper extends StagingHelper
{
    private static Log logger =
        LogFactory.getLog(WSStagingHelper.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static PerformanceLog performanceLogger = new PerformanceLog(
        WSStagingHelper.class.getName() + ".performance");

 
    public EndpointReferenceType submitStagingRequest(
        ManagedExecutableJobResource        resource,
        BaseRequestType                     stagingRequest)
        throws                              Exception
    {
        logger.trace(getLoggingPrefix(resource) +
            "Entering sumbitStagingRequest()");

        //setup the RFT factory stub
        ReliableFileTransferFactoryServiceAddressingLocator factoryLocator
            = new ReliableFileTransferFactoryServiceAddressingLocator();
        ReliableFileTransferFactoryPortType factoryPort
            = factoryLocator.getReliableFileTransferFactoryPortTypePort(
                resource.getReliableFileTransferFactoryEndpoint());
        ManagedExecutableJobHome.getHome().enableLocalTransport(
            (Stub) factoryPort);
        setStagingStubSecurity(resource, (Stub) factoryPort);

        //setup the create input
        CreateReliableFileTransferInputType createInput
            = new CreateReliableFileTransferInputType();
        if (stagingRequest instanceof TransferRequestType) {
            createInput.setTransferRequest((TransferRequestType)stagingRequest);
        } else {
            createInput.setDeleteRequest((DeleteRequestType)stagingRequest);
        }
        Calendar terminationTime = Calendar.getInstance();
        terminationTime.add(Calendar.HOUR, 12);
        createInput.setInitialTerminationTime(terminationTime);

        //create the transfer
        logger.trace(getLoggingPrefix(resource) +
            "calling RFTF.createReliableFileTransfer()");
        CreateReliableFileTransferOutputType createOutput
            = factoryPort.createReliableFileTransfer(createInput);
        EndpointReferenceType transferJobEndpoint
            = createOutput.getReliableTransferEPR();
        if (logger.isDebugEnabled())
        {
            logger.debug("RFT service URL: "
                + transferJobEndpoint.getAddress().toString());
        }

        //start the transfer and setup the notifications
        StartOutputType startResponse
            = startTransferJob(resource, transferJobEndpoint);

        logger.trace(getLoggingPrefix(resource) +
            "Leaving sumbitStagingRequest()");
        
        return transferJobEndpoint;
    }

    public StartOutputType startTransferJob(
        ManagedExecutableJobResource            resource,
        EndpointReferenceType                   transferJobEndpoint)
        throws                                  Exception
    {
        logger.trace(getLoggingPrefix(resource) +
            "Entering startTransferJob()");

        //register with the staging listener
        StagingListener.registerTransferJob(
            transferJobEndpoint, (ResourceKey) resource.getID());

        //setup the transfer stub
        ReliableFileTransferServiceAddressingLocator transferLocator
            = new ReliableFileTransferServiceAddressingLocator();
        ReliableFileTransferPortType transferJobPort
            = transferLocator.getReliableFileTransferPortTypePort(
                transferJobEndpoint);
        ManagedExecutableJobHome.getHome().enableLocalTransport(
            (Stub) transferJobPort);
        setStagingStubSecurity(resource, (Stub) transferJobPort);

        //setup the RFT notification listener.
        NotificationConsumerManager notificationConsumerManager
            = ManagedExecutableJobHome.getNotificationConsumerManager();
        List topicPath = new LinkedList();
        //topicPath.add(RFTConstants.OVERALL_STATUS_RESOURCE);
        topicPath.add(RFTConstants.REQUEST_STATUS_RESOURCE);

        ResourceSecurityDescriptor securityDescriptor
            = new ResourceSecurityDescriptor();

        securityDescriptor.setAuthz(Authorization.AUTHZ_NONE);

        Vector authMethod = new Vector();
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String wsTransportProtocol
            = factoryHomeConfiguration.getStagingProtocol();
        if (wsTransportProtocol == null)
        {
            wsTransportProtocol = ServiceHost.getDefaultProtocol();
        }
        if (wsTransportProtocol.equals("http"))
        {
            authMethod.add(GSISecureMsgAuthMethod.BOTH);
        } else
        {
            authMethod.add(GSITransportAuthMethod.BOTH);
        }
        securityDescriptor.setAuthMethods(authMethod);

        notificationConsumerManager.startListening();
        EndpointReferenceType notificationConsumerEndpoint =
            notificationConsumerManager.createNotificationConsumer(
                topicPath,
                StagingListener.getInstance(),
                securityDescriptor);
        resource.setResourceDatum(
            ManagedJobConstants.RD_NOTIFICATION_CONSUMER_ENDPOINT,
            notificationConsumerEndpoint);

        Subscribe subscriptionRequest = new Subscribe();
        String resourceHandle = null;
        subscriptionRequest.setConsumerReference(notificationConsumerEndpoint);
        TopicExpressionType topicExpression = new TopicExpressionType(
            WSNConstants.SIMPLE_TOPIC_DIALECT,
            RFTConstants.REQUEST_STATUS_RESOURCE);
            //RFTConstants.OVERALL_STATUS_RESOURCE);
        subscriptionRequest.setTopicExpression(topicExpression);

        SubscribeResponse subscribeResponse
            = transferJobPort.subscribe(subscriptionRequest);
        EndpointReferenceType subscriptionEndpoint
            = subscribeResponse.getSubscriptionReference();
        resource.setResourceDatum(
            ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT,
            subscriptionEndpoint);

        //start transfer
        logger.trace(getLoggingPrefix(resource) + "calling RFT.start()");
        StartOutputType startResponse = transferJobPort.start(new Start());

        logger.trace(getLoggingPrefix(resource) +
            "Leaving startTransferJob()");

        return startResponse;
    }

    public void destroyTransferJob(
        ManagedExecutableJobResource        resource,
        EndpointReferenceType               transferJobEndpoint)
        throws                              Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) +
                "Entering destroyTransferJob()");
        }

        //destroy status notification subscription
        EndpointReferenceType subscriptionEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT);
        if (subscriptionEndpoint != null)
        {
            if (logger.isDebugEnabled()) {
                logger.debug(getLoggingPrefix(resource) +
                    "Destroying transfer notification subscription.");
            }
            SubscriptionManager subscriptionPort
                = new WSBaseNotificationServiceAddressingLocator().
                    getSubscriptionManagerPort(subscriptionEndpoint);
            ManagedExecutableJobHome.getHome().enableLocalTransport(
                (Stub) subscriptionPort);
            setStagingStubSecurity(resource, (Stub) subscriptionPort);

            resource.setResourceDatum(
               ManagedExecutableJobConstants.RD_SUBSCRIPTION_ENDPOINT,
               null);

            subscriptionPort.destroy(new Destroy());
        }

        if (transferJobEndpoint != null)
        {
            //unregister the transfer from the staging listener
            if (logger.isDebugEnabled()) {
                logger.debug(getLoggingPrefix(resource) +
                    "Unregister transfer from staging listener.");
            }
            StagingListener.unregisterTransferJob(transferJobEndpoint);

            //destroy the transfer
            if (logger.isDebugEnabled()) {
                logger.debug(getLoggingPrefix(resource) +
                    "Destroying transfer.");
            }
            ReliableFileTransferServiceAddressingLocator transferLocator
                = new ReliableFileTransferServiceAddressingLocator();
            ReliableFileTransferPortType transferJobPort
                = transferLocator.getReliableFileTransferPortTypePort(
                    transferJobEndpoint);
            ManagedExecutableJobHome.getHome().enableLocalTransport(
                (Stub) transferJobPort);
            setStagingStubSecurity(resource, (Stub) transferJobPort);

            transferJobPort.destroy(new Destroy());
        }

        //destroy notification consumer
        EndpointReferenceType notificationConsumerEndpoint
            = (EndpointReferenceType) resource.getResourceDatum(
                ManagedJobConstants.RD_NOTIFICATION_CONSUMER_ENDPOINT);
        if (notificationConsumerEndpoint != null)
        {
            if (logger.isDebugEnabled()) {
                logger.debug(getLoggingPrefix(resource) +
                    "Destroying transfer notification consumer.");
            }
            NotificationConsumerManager notificationConsumerManager
                = ManagedExecutableJobHome.getNotificationConsumerManager();
            notificationConsumerManager.startListening();
            notificationConsumerManager.removeNotificationConsumer(
                notificationConsumerEndpoint);
        }

        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) +
                "Leaving destroyTransferJob()");
        }
    }

    public RequestStatusType poll(
        ManagedExecutableJobResource        resource,
        EndpointReferenceType               transferJobEndpoint)
    {
        try {

            if (transferJobEndpoint != null)
            {
                ReliableFileTransferServiceAddressingLocator transferLocator
                    = new ReliableFileTransferServiceAddressingLocator();
                ReliableFileTransferPortType transferJobPort
                    = transferLocator.getReliableFileTransferPortTypePort(
                        transferJobEndpoint);
                ManagedExecutableJobHome.getHome().enableLocalTransport(
                    (Stub) transferJobPort);
                setStagingStubSecurity(resource, (Stub) transferJobPort);

                if (logger.isDebugEnabled()) {
                    logger.debug(getLoggingPrefix(resource) +
                        "polling for transfer job status.");
                }

                GetResourcePropertyResponse response
                    = transferJobPort.getResourceProperty(
                        RFTConstants.REQUEST_STATUS_RESOURCE);
                        //RFTConstants.OVERALL_STATUS_RESOURCE);

                SOAPElement [] any = response.get_any();
                RequestStatusType status =
                    (RequestStatusType) ObjectDeserializer.
                    toObject(any[0], RequestStatusType.class);

                return status;
            }
        } catch (Exception e)
        {
            logger.debug("unable to poll for transfer job status", e);
        }

        return null;
    }


    private void setStagingStubSecurity(
        ManagedExecutableJobResource        resource,
        Stub                                stub)
        throws                              RemoteException
    {
        //determine security type (message-level or transport-level)
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String wsTransportProtocol
            = factoryHomeConfiguration.getStagingProtocol();
        String securityType = Constants.GSI_TRANSPORT;
        if (wsTransportProtocol == null)
        {
            //wsTransportProtocol = ServiceHost.getDefaultProtocol();
            wsTransportProtocol = "local";
        }
        if (logger.isDebugEnabled())
        {
            logger.debug(getLoggingPrefix(resource) +
                "Transport Protocol: " + wsTransportProtocol);
        }
        //if (wsTransportProtocol.equals("http"))
        if (   wsTransportProtocol.equals("http")
            || wsTransportProtocol.equals("local"))
        {

            securityType = Constants.GSI_SEC_MSG;
        }

        ClientSecurityDescriptor secDesc = new ClientSecurityDescriptor();

        //set security type
        if (securityType.equals(Constants.GSI_SEC_MSG))
        {
            secDesc.setGSISecureMsg(Constants.SIGNATURE);
        } else
        {
            secDesc.setGSITransport(Constants.SIGNATURE);
        }

        DelegatedCredential stagingCredential = null;
        try
        {
            stagingCredential = resource.getStagingCredential();
        }
         catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                Resources.CREDENTIAL_ACQUISITION_ERROR);
            FaultType fault = FaultUtils.makeFault(
                FaultType.class,
                (ResourceKey) resource.getID(),
                errorMessage,
                e,
                resource.getState(),
                null,
                0);
            resource.setFault(fault);
        }
        if (stagingCredential == null)
        {
            String errorMessage
                = i18n.getMessage(Resources.MISSING_STAGING_CREDENTIAL);
            throw new RemoteException(errorMessage);
        }
        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) +
                "Delegated Staging Credential Object: "
                + stagingCredential.toString());
            try {
                logger.debug(getLoggingPrefix(resource) + 
                    "Using proxy with subject "
                    +stagingCredential.getCredential().getIdentity()
                    +"to contact the staging service");
            } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                    Resources.DELEGATED_PROXY_SUBJECT_ERROR);
                logger.error(errorMessage);
            }
        }
        try {
            secDesc.setGSSCredential(
                new GlobusGSSCredentialImpl(
                    stagingCredential.getCredential(),
                    GSSCredential.INITIATE_AND_ACCEPT));
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.GLOBUS_CREDENTIAL_TO_GSS_CONVERSION_ERROR);
            throw new RemoteException(errorMessage, e);
        }

        //set factory authorization method
        String stagingSubject = factoryHomeConfiguration.getStagingSubject();
        if (logger.isDebugEnabled()) {
            logger.debug(getLoggingPrefix(resource) +
                "Using RFT Authorization subject " + stagingSubject);
        }
        logger.debug(getLoggingPrefix(resource) +
            "Staging Subject: -->" + stagingSubject + "<--");
        if ((stagingSubject != null) && !stagingSubject.equals(""))
        {
            secDesc.setAuthz(new IdentityAuthorization(stagingSubject));
        }
        else
        {
            ContainerSecurityConfig containerSecurityConfig = null;
            ContainerSecurityDescriptor containerSecurityDescriptor = null;
            Subject containerSubject = null;
            Set principals = null;

            try
            {
                containerSecurityConfig
                    = ContainerSecurityConfig.getConfig();
                containerSecurityDescriptor
                    = containerSecurityConfig.getSecurityDescriptor();
                containerSubject
                    = containerSecurityDescriptor.getSubject();
                if (containerSubject != null)
                {
                    principals = containerSubject.getPrincipals();
                }
            } catch (Exception e)
            {
                String errorMessage = i18n.getMessage(
                    Resources.CREDENTIAL_CONFIGURATION_ERROR);
                throw new RuntimeException(errorMessage);
            }

            if (containerSubject != null)
            {

                if (containerSecurityConfig.getSecurityDescriptorFile() == null)
                {
                    //using default creds, assume identity authz
                    if (logger.isDebugEnabled()) {
                        logger.debug("Using default creds. "
                                    +"Assume identity authz");
                    }
                    secDesc.setAuthz(new IdentityAuthorization(
                        principals.iterator().next().toString()));
                }
                else
                if (containerSecurityDescriptor.getProxyFilename() != null)
                {
                    logger.debug("Found proxy file configuration");

                    secDesc.setAuthz(new IdentityAuthorization(
                        principals.iterator().next().toString()));
                }
                else
                if (   (containerSecurityDescriptor.getCertFilename() != null)
                    && (containerSecurityDescriptor.getKeyFilename() != null))
                {
                    logger.debug("Found host and cert configuration.");

                    secDesc.setAuthz(HostAuthorization.getInstance());
                }
                else
                {
                    logger.debug("No credentials specified, assuming user creds");

                    secDesc.setAuthz(new IdentityAuthorization(
                        principals.iterator().next().toString()));
                }
            } else {
                //no container credentials

                String errorMessage = i18n.getMessage(
                    Resources.CREDENTIAL_CONFIGURATION_ERROR);
                throw new RuntimeException(errorMessage);
            }
        }

        stub._setProperty(Constants.CLIENT_DESCRIPTOR,
                          secDesc);
    }

}
