/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.factory;

import java.util.StringTokenizer;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.xml.namespace.QName;

import org.apache.axis.MessageContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

import org.globus.delegation.DelegationConstants;
import org.globus.delegation.DelegationUtil;
import org.globus.exec.utils.Resources;
import org.globus.exec.generated.ManagedJobFactoryResourceProperties;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.service.exec.ManagedExecutableJobHome;
import org.globus.exec.service.utils.AuthorizationHelper;
import org.globus.exec.service.utils.ConfigurationHelper;
import org.globus.exec.utils.service.ManagedJobHelper;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.ContextUtils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.ResourceContextException;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceKey;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.globus.wsrf.impl.servicegroup.client.MDSConfiguration;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;

public class ManagedJobFactoryHome extends ResourceHomeImpl
{
    private static Log logger =
        LogFactory.getLog(ManagedJobFactoryHome.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static FactoryHomeConfiguration CONFIGURATION;

    private boolean registerToMDS = false;
    private boolean initialized = false;

    public ManagedJobFactoryHome() {
        //DO NOT DO ANYTHING HERE. PUT IT IN initialize().
    }

    public synchronized void initialize() throws Exception
    {
        if (this.initialized)
        {
            return;
        }


        //initialize the service path
        if (ManagedJobFactoryService.SERVICE_PATH == null)
        {
            MessageContext context = MessageContext.getCurrentContext();
            if (context == null)
            {
                throw new RuntimeException("No Context!");
            }

            ManagedJobFactoryService.SERVICE_PATH
                = ContextUtils.getTargetServicePath(context);
                    
        }

        //prime this while we have a context attached to this thread
        AuthorizationHelper.initialize();

        //necessary for deserialization
        this.keyTypeName = ManagedJobFactoryConstants.RESOURCE_KEY_QNAME;

        super.initialize();
        logger.debug("Creating ManagedJobFactoryHome");

        ResourceContext ctx = ResourceContext.getResourceContext();
        Context initialContext = new InitialContext();
        //getting common configuration values
        CONFIGURATION = (FactoryHomeConfiguration)
            ConfigurationHelper.getResourceHomeConfiguration();

        // load the MDS config before creating each factory resource
        try {
            // read config
            String config = org.globus.wsrf.Constants.JNDI_SERVICES_BASE_NAME +
                ResourceContext.getResourceContext().getService()
                + "/mdsConfiguration";
            MDSConfiguration mdsConfiguration =
                (MDSConfiguration) initialContext.lookup(config);

            // determine if we should register
            registerToMDS= mdsConfiguration.getRegAsBoolean();
            logger.debug("Register to MDS: "+registerToMDS);
                
        } catch(Exception e) {
            logger.debug("Couldn't load MDS config: "+e);
        }

        //getting configuration for each LRM. Each LRM JNDI resource
        //must be of a type equal to or derived from
        //FactoryResourceConfiguration
        NamingEnumeration objects = initialContext.listBindings(
            ConfigurationHelper.getServiceJNDILocation());
        while (objects.hasMore())
        {
            Binding binding = (Binding)objects.next();
            logger.debug("JNDI: binding found " + binding);
            Object boundJNDIObject = binding.getObject();
            if (boundJNDIObject != null) {
                Class boundJNDIObjectClazz = boundJNDIObject.getClass();
                if (Reference.class.isAssignableFrom(boundJNDIObjectClazz)) {
                    boundJNDIObjectClazz = Class.forName(
                        ((Reference)boundJNDIObject).getClassName());
                }
                logger.debug("boundJNDIObjectClazz=" + boundJNDIObjectClazz);
                if (FactoryResourceConfiguration.class.isAssignableFrom(
                    boundJNDIObjectClazz))
                {
                    FactoryResourceConfiguration lrmConfig =
                        (FactoryResourceConfiguration)ConfigurationHelper.
                        getJNDIObjectForCurrentService(binding.getName());

                    String strLRM = lrmConfig.getLocalResourceManagerName();
                    logger.debug("Creating factory resource for " + strLRM);
                    /*TODO: move Condor-specific config from the
                      Home configuration*/

                    //get scratch directory
                    String scratchDirectory = lrmConfig.getScratchDirectory();

                    //get substitution variable definitions file path
                    String subDefFile
                        = lrmConfig.getSubstitutionDefinitionsFile();

                    //get substitution variable definitions refresh period
                    String subDefRefreshPeriod
                        = lrmConfig.getSubstitutionDefinitionsRefreshPeriod();

                    String enableDefaultSoftwareEnvironment
                        = lrmConfig.getEnableDefaultSoftwareEnvironment();

                    create(
                        strLRM,
                        scratchDirectory,
                        subDefFile,
                        subDefRefreshPeriod,
                        enableDefaultSoftwareEnvironment);
                }
            }
        }

        
        this.initialized = true;
    }

    /**
     * This function is private as we do not want to let other objects create
     * factory resources.
     * @param localResourceManager the local resource manager to manage the
     *        new job with.
     * @throws Exception
     * @return ManagedJobFactoryResource
     */
    private ManagedJobFactoryResource create(
        String                              localResourceManager,
        String                              scratchDirectory,
        String                              subDefFile,
        String                              subDefRefreshPeriod,
        String                              enableDefaultSoftwareEnvironment)
        throws                              Exception
    {
        SimpleResourceKey key
            = ManagedJobHelper.getResourceKey(localResourceManager);
        ManagedJobFactoryResource factoryResource =
            new ManagedJobFactoryResource();
        factoryResource.initialize(
            createResourceBean(localResourceManager, scratchDirectory),
            ManagedJobFactoryConstants.RESOURCE_PROPERTY_SET,
            key,
            subDefFile,
            subDefRefreshPeriod,
            enableDefaultSoftwareEnvironment);
        this.add(key, factoryResource); //or put the hashCode of the stringKey and override find()
        return factoryResource;
    }


    /** Override add to optionally publish to MDS, if so configured. */
    protected void add(ResourceKey key, Resource resource)
    {
        logger.debug("In custom add");
        // pass the call to the superclass
        super.add(key,resource);

        // skip the rest of this method if we aren't registering to MDS
        if(!registerToMDS) return;

        logger.debug("registering to MDS");

        ResourceContext ctx;
        try {
            ctx = ResourceContext.getResourceContext();
        } catch(ResourceContextException e) {
            logger.error("Could not get ResourceContext: "+e);
            return;
        }

        EndpointReferenceType epr;
        try {
            epr = AddressingUtils.createEndpointReference(ctx, key);
        } catch(Exception e) {
            logger.error("Could not form EPR: "+e);
            return;
        }

        try {
            // perform registraion
            ServiceGroupRegistrationClient.register(
                                 ctx,
                                 epr,
                                 "/etc/gram-service/registration.xml");
        } catch(Exception e) {
            logger.error("Exception when trying to register to MDS: "
                         +e);
        }
    }






    private ManagedJobFactoryResourceProperties createResourceBean(
        String localResourceManager, String scratchDirectory)
    {

        ManagedJobFactoryResourceProperties resourceProperties =
            new ManagedJobFactoryResourceProperties();

        //Set initial values of Resource Properties:

        resourceProperties.setLocalResourceManager(localResourceManager);
        resourceProperties.setScratchBaseDirectory(scratchDirectory);

        resourceProperties.setGlobusLocation(
            CONFIGURATION.getGlobusLocation());
        resourceProperties.setHostCPUType(
            CONFIGURATION.getHostCpuType());
        resourceProperties.setHostManufacturer(
            CONFIGURATION.getHostManufacturer());
        resourceProperties.setHostOSName(
            CONFIGURATION.getHostOsName());
        resourceProperties.setHostOSVersion(
            CONFIGURATION.getHostOsVersion());
        resourceProperties.setCondorArchitecture(
            CONFIGURATION.getCondorArchitecture());
        resourceProperties.setCondorOS(
            CONFIGURATION.getCondorOS());

        return resourceProperties;
    }

    public static ManagedJobFactoryHome getHome() throws NamingException {
        logger.trace("getHome() called");
        ManagedJobFactoryHome home;

        home = (ManagedJobFactoryHome)ConfigurationHelper.getResourceHome(
            ManagedJobFactoryConstants.SERVICE_PATH_WITHOUT_CONTEXT);

        logger.trace("getHome() returning " +  (
                (home == null) ? "null" : home.toString()));
        return home;
    }

    public static FactoryHomeConfiguration getConfiguration() {
        return CONFIGURATION;
    }
}
