/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.factory;

import java.rmi.RemoteException;
import java.util.Calendar;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.oasis.wsn.Subscribe;
import org.oasis.wsrf.faults.BaseFaultType;
import org.oasis.wsrf.lifetime.UnableToSetTerminationTimeFaultType;
import org.globus.exec.generated.CreateManagedJobInputType;
import org.globus.exec.generated.CreateManagedJobOutputType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.MultiJobDescriptionType;
import org.globus.exec.service.job.ManagedJobHome;
import org.globus.exec.service.job.ManagedJobResource;
import org.globus.exec.service.utils.ConfigurationHelper;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.util.I18n;
import org.globus.wsrf.Constants;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.notification.SubscribeHelper;
import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.utils.PerformanceLog;

public class ManagedJobFactoryService
{
    private static PerformanceLog performanceLogger = new PerformanceLog(
        ManagedJobFactoryService.class.getName() + ".performance");

    private static Log logger =
        LogFactory.getLog(ManagedJobFactoryService.class.getName());

    static I18n i18n = I18n.getI18n(
        org.globus.exec.utils.Resources.class.getName());

    private static FactoryServiceConfiguration SERVICE_CONFIGURATION = null;

    static String SERVICE_PATH = null;

    public ManagedJobFactoryService() {
        logger.trace("Creating ManagedJobFactoryService " + this);
    }

    public CreateManagedJobOutputType createManagedJob(
            CreateManagedJobInputType               jobCreationInput)
            throws                                  RemoteException,
                                                    BaseFaultType
    {
        if (logger.isDebugEnabled())
        {
            logger.debug("Entering createManagedJob()");
            performanceLogger.start();
            /*
            logger.debug("<startTime name=\"createManagedJob\">"
                       +System.currentTimeMillis()
                       +"</startTime>");
                       */
        }
        
        //========= Initial Termination Time

        Calendar requestedTermTime =
            jobCreationInput.getInitialTerminationTime();

        //is there a piece of code we can reuse here? Maybe not.
        //Then we should make this available in the framework. How?
        Calendar currentTime = Calendar.getInstance();
        Calendar initTermTime = currentTime;
        if (requestedTermTime == null) {
            //TODO test if this branch executes for nil AND absent
            initTermTime = null;
            if (logger.isDebugEnabled()) {
                logger.debug(
                "setting job resource termination time to null i.e. infinite");
            }
        } else {
            // if termTime in the past - no creation and throw Fault
            if (logger.isDebugEnabled()) {
                logger.debug("requested termination time: "
                            + requestedTermTime.getTime());
                logger.debug("current time: " + currentTime.getTime());
            }
            if (requestedTermTime.getTime().before(currentTime.getTime())) {
                throw new UnableToSetTerminationTimeFaultType();
            } else {
                initTermTime = requestedTermTime;
                if (logger.isDebugEnabled()) {
                    logger.debug(
                    "setting job resource termination time to requested time");
                }
            }
        }
        //TODO test the setResourceProperty(TermTime) cannot work!

        CreateManagedJobOutputType response = new CreateManagedJobOutputType();
        response.setCurrentTime(currentTime);
        response.setNewTerminationTime(initTermTime);

        //TODO Save and associate credential

        //========== Examination/Transformation of RSL
        //TODO: check input RSL is not empty, and valid.
        //Do transformations if needed
        //(for instance for MyJob/DUROC via a plug-in?)

        JobDescriptionType jobDescription = jobCreationInput.getJob();
        if (jobDescription == null)
        {
            jobDescription = jobCreationInput.getMultiJob();
        }
        if (jobDescription == null)
        {
            throw new RemoteException(i18n.getMessage("rslMissing"));
        }

        //clean off some serialization remnants
        try
        {
            jobDescription = (JobDescriptionType) ObjectSerializer.clone(
                jobDescription);
        }
        catch (Exception e)
        {
            throw new RemoteException("bad RSL", e);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("received RSL : \n" +
                         RSLHelper.convertToString(jobDescription));
        }

        EndpointReferenceType epr = new EndpointReferenceType();
        EndpointReferenceType subscriptionEPR = null;
        ResourceHomeImpl jobHome = null;

        try
        {
            ResourceContext ctx = ResourceContext.getResourceContext();
            ManagedJobFactoryResource factoryResource = null;
            try
            {
                //get localResourceManager from current resource
                factoryResource = (ManagedJobFactoryResource) ctx.getResource();
            }
            catch (NoSuchResourceException nsre)
            {
                String serviceAddress = ctx.getServiceURL().toString();
                String resourceKey = ctx.getResourceKey().getValue().toString();
                throw new RemoteException(
                    i18n.getMessage(
                        Resources.BAD_FACTORY_RESOURCE,
                        new String[] { serviceAddress, resourceKey}));
            }

            String localResourceManager = (String)factoryResource.
                getResourcePropertySet().get(
                   ManagedJobFactoryConstants.RP_LOCAL_RESOURCE_MANAGER).get(0);

            //get the caller's user subject
            SecurityManager securityManager = SecurityManager.getManager();
            String userSubject = securityManager.getCaller();

            //get the caller's local user ID via system grid-mapfile lookup
            String servicePath = ctx.getService();

            logger.debug("servicePath=" + servicePath);
            String localUserId = null;
            String[] localUserIds = securityManager.getLocalUsernames();
            String requestedLocalUserId = jobDescription.getLocalUserId();
            if (requestedLocalUserId != null) {
                boolean foundLocalUserId = false;
                for (int index=0;
                     (index<localUserIds.length) && !foundLocalUserId;
                     index++)
                {
                    if (requestedLocalUserId.equals(localUserIds[index])) {
                        foundLocalUserId = true;
                    }
                }

                if (foundLocalUserId) {
                    localUserId = requestedLocalUserId;
                } else {
                    throw new RemoteException(
                        i18n.getMessage(
                            "unauthorizedLocalUserId",
                            new Object[] {requestedLocalUserId}));
                }
            } else {
                localUserId = localUserIds[0];
            }

            if (logger.isDebugEnabled()) {
                logger.debug("createManagedJob() called..."
                            + "\nCaller ID: " + userSubject
                            + "\nLooked up local user ID: " + localUserId);
            }

            if (logger.isDebugEnabled()) {
                if (jobDescription instanceof MultiJobDescriptionType) {
                    if (logger.isDebugEnabled()) {
                        logger.debug("received a multi-job RSL");
                    }
                } else {
                    if (logger.isDebugEnabled()) {
                        logger.debug("received a single-job RSL");
                    }
                }
            }

            /*TODO: should make job creation more general to support
             *      other types of jobs besides executable and multi
             */

            FactoryServiceConfiguration serviceConfiguration
                = getServiceConfiguration();
            String execJobServicePath
                = serviceConfiguration.getManagedExecutableJobServicePath();
            ManagedJobHome execJobHome
                = (ManagedJobHome)ConfigurationHelper
                .getResourceHome(execJobServicePath);
            String multiJobServicePath
                = serviceConfiguration.getManagedMultiJobServicePath();
            ManagedJobHome multiJobHome
                = (ManagedJobHome)ConfigurationHelper
                .getResourceHome(multiJobServicePath);

            /*
            int jobResourceCount
                = execJobHome.getResourceCount()
                + multiJobHome.getResourceCount();
            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String maxJobLimit = factoryHomeConfiguration.getMaxJobLimit();
            if (jobResourceCount >= new Integer(maxJobLimit).intValue())
            {
                String errorMessage = i18n.getMessage(
                    Resources.MAX_JOB_RESOURCE_LIMIT_REACHED,
                    maxJobLimit);
                throw new RemoteException(errorMessage);
            }
            */

            String jobServicePath = null;
            if (localResourceManager.toLowerCase().equals("multi")) {
                //get multi-job service path
                jobServicePath = multiJobServicePath;
                jobHome = (ResourceHomeImpl) multiJobHome;
            } else {
                //get executable job service path
                jobServicePath = execJobServicePath;
                jobHome = (ResourceHomeImpl) execJobHome;
            }

            if (logger.isDebugEnabled()) {
                logger.debug("Managed Job Service Path: "
                            + jobServicePath);
            }

            AttributedURI idempotenceIdUri = jobCreationInput.getJobID();
            String idempotenceId = null;
            if (idempotenceIdUri != null) {
                idempotenceId = idempotenceIdUri.getSchemeSpecificPart();
            }

            String serviceURL = ServiceHost.getBaseURL() + jobServicePath;

            //performanceLogger.start();
            if (logger.isDebugEnabled())
            {
                logger.debug("at create()");
            }

            ResourceKey key = null;
            ManagedJobResource resource = null;

            synchronized (this)
            {
                key = ((ManagedJobHome)jobHome).create(
                    initTermTime,
                    currentTime,
                    localResourceManager,
                    jobDescription,
                    userSubject,
                    localUserId,
                    idempotenceId,
                    serviceURL);

                resource = (ManagedJobResource)jobHome.find(key);

                //TODO store as RP "ResourceEndpointReference"
                epr = AddressingUtils.createEndpointReference(
                    serviceURL, key);

                if (logger.isDebugEnabled())
                {
                    logger.debug("at find()");
                }
            }

            //subscribe the creator for notifications if so desired
            //performanceLogger.start();
            Subscribe subscription = jobCreationInput.getSubscribe();
            if (subscription != null)
            {
                SubscribeHelper subscribeHelper = new SubscribeHelper(
                    ctx, resource, key,
                    Constants.JNDI_SERVICES_BASE_NAME
                        + jobServicePath + Constants.HOME_NAME,
                    subscription);
                if (logger.isDebugEnabled())
                {
                    logger.debug("at subscribe()");
                }
                subscriptionEPR = subscribeHelper.subscribe();
            }
            //performanceLogger.stop("subscribeHelper.subscribe()");

            //start the state machine
            //performanceLogger.start();
            if (logger.isDebugEnabled())
            {
                logger.debug("at start()");
            }

            resource.start();
            //performanceLogger.stop("resource.start()");
        }
        catch (Exception e) {
            String errorMessage = i18n.getMessage("jobCreationFailed");
            logger.error(errorMessage, e);
            throw new RemoteException(errorMessage, e);
        }

        response.setSubscriptionEndpoint(subscriptionEPR);
        response.setManagedJobEndpoint(epr);

        //TODO: record response for that messageID, for reliability
        
        if (logger.isDebugEnabled())
        {
            logger.debug("Leaving createManagedJob()");
            performanceLogger.stop("createManagedJob");
            /*
            logger.debug("<endTime name=\"createManagedJob\">"
                       +System.currentTimeMillis()
                       +"</endTime>");
                       */
        }

        return response;
    }

    public synchronized static FactoryServiceConfiguration getServiceConfiguration()
    {
        if (ManagedJobFactoryService.SERVICE_CONFIGURATION == null)
        {
            try
            {
                ManagedJobFactoryService.SERVICE_CONFIGURATION
                    = (FactoryServiceConfiguration)
                    ConfigurationHelper.getServiceConfiguration(
                       ManagedJobFactoryConstants.SERVICE_NAME);
            }
            catch (Exception e)
            {
                String errorMessage = i18n.getMessage(
                    Resources.JNDI_MJFS_CONFIG_ERROR);
                logger.error(errorMessage, e);
                throw new RuntimeException(errorMessage, e);
            }
        }

        return ManagedJobFactoryService.SERVICE_CONFIGURATION;
    }

    public static String getServicePath()
    {
        return ManagedJobFactoryService.SERVICE_PATH;
    }
}
