/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.factory;

import java.io.FileInputStream;
import java.net.URL;
import java.util.Date;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;

import org.apache.axis.utils.XMLUtils;

import org.apache.axis.message.addressing.EndpointReferenceType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

import org.globus.exec.utils.Resources;
import org.globus.exec.utils.ManagedJobFactoryConstants;

import org.globus.delegation.DelegationConstants;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;
import org.globus.wsrf.utils.AddressingUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class StagingDelegationFactoryEndpointResourceProperty
    extends                                 ReflectionResourceProperty
{

    private static Log logger = LogFactory.getLog(
        StagingDelegationFactoryEndpointResourceProperty.class.getName());

    private static StagingDelegationFactoryEndpointResourceProperty
        resourceProperty = null;

    private EndpointReferenceType endpoint = null;

    public static StagingDelegationFactoryEndpointResourceProperty getInstance()
        throws                              Exception
    {
        if (StagingDelegationFactoryEndpointResourceProperty.resourceProperty
            == null)
        {
            StagingDelegationFactoryEndpointResourceProperty.resourceProperty
                = new StagingDelegationFactoryEndpointResourceProperty();
        }

        return StagingDelegationFactoryEndpointResourceProperty
               .resourceProperty;
    }

    /**
    @throws java.lang.Exception passes along any exception encountered
        during construction of underlying ReflectionResourceProperty.
    */
    private StagingDelegationFactoryEndpointResourceProperty()
        throws                              Exception
    {
        super(new SimpleResourcePropertyMetaData(
            ManagedJobFactoryConstants.RP_STAGING_DELEGATION_FACTORY_ENDPOINT));
        setObject(this);
        setPropertyName("DynamicValue");
        initialize();
    }

    /* Getter method for property "DynamicValue" determined through reflection.
     */
    public EndpointReferenceType getDynamicValue()
    {
        if (this.endpoint == null)
        {
            URL serviceUrl = null;
            try
            {
                URL defaultBaseUrl = ServiceHost.getBaseURL();
                String protocol = defaultBaseUrl.getProtocol();
                String host = defaultBaseUrl.getHost();
                int port = defaultBaseUrl.getPort();
                String path = defaultBaseUrl.getPath();
                String name = DelegationConstants.FACTORY_PATH;

                //check for configured overrides
                FactoryHomeConfiguration factoryHomeConfiguration
                    = ManagedJobFactoryHome.getConfiguration();
                String protocolOverride
                    = factoryHomeConfiguration.getStagingProtocol();
                if (protocolOverride != null)
                {
                    protocol = protocolOverride;
                }
                String hostOverride
                    = factoryHomeConfiguration.getStagingHost();
                if (hostOverride != null)
                {
                    host = hostOverride;
                }
                String portOverride
                    = factoryHomeConfiguration.getStagingPort();
                if (portOverride != null)
                {
                    port = Integer.parseInt(portOverride);
                }
                String pathOverride
                    = factoryHomeConfiguration.getStagingServicePath();
                if (pathOverride != null)
                {
                    path = pathOverride;
                }
                String nameOverride
                    = factoryHomeConfiguration
                    .getStagingDelegationFactoryName();
                if (nameOverride != null)
                {
                    name = nameOverride;
                }

                serviceUrl = new URL(
                    protocol,
                    host,
                    port,
                    path + name);


                if (logger.isDebugEnabled())
                {
                    logger.debug("Delegation Servcie Address: " + serviceUrl);
                }
                this.endpoint = AddressingUtils.createEndpointReference(
                    serviceUrl.toString(), null);
            } catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.DELEGATION_EPR_CREATION_ERROR,
                                    serviceUrl);
                throw new RuntimeException(errorMessage, e);
            }
        }

        return this.endpoint;
    }

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
}

