/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.multi;

import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.NonNegativeInteger;

import org.globus.exec.generated.ManagedMultiJobResourcePropertiesType;
import org.globus.exec.generated.InternalStateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ServiceLevelAgreementType;
import org.globus.exec.generated.ManagedMultiJobResourceDataType;
import org.globus.exec.generated.ManagedMultiJobResourceStateType;
import org.globus.exec.generated.MultiJobDescriptionType;
import org.globus.exec.generated.NameValuePairType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.monitoring.JobStateChangeListener;
import org.globus.exec.service.factory.FactoryServiceConfiguration;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.job.ManagedJobHome;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.ManagedMultiJobConstants;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.exec.utils.service.ManagedJobHelper;
//import org.globus.resolver.ResolverUtils;
import org.globus.util.I18n;
import org.globus.wsrf.Constants;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceIdentifier;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.container.ServiceContainer;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.ReflectionResource;
import org.globus.wsrf.PersistentResource;
import org.globus.wsrf.utils.PerformanceLog;
import org.globus.wsrf.utils.XmlPersistenceHelper;
import org.globus.wsrf.utils.AddressingUtils;

public class ManagedMultiJobHome extends    ResourceHomeImpl
                                 implements ManagedJobHome {

    private boolean recovered = false;

    public ManagedMultiJobHome() {
        //DO NOT DO ANYTHING HERE. PUT IT IN initialize().
    }

    public synchronized void initialize() throws Exception {
        //Key deserialization configuration (easier here than in JNDI config)
        this.keyTypeName = ManagedJobConstants.RESOURCE_KEY_QNAME;

        super.initialize();
    }

    /** This method assures that any jobs which were being processed by the
     *  state machine can continue to do so after container recovery.
     *  Subscriptions to the JobStateMonitor are also restored.
     *  It is intended to be called by ManagedJobFactoryResource after starting
     *  the JobStateMonitor.
     */
    public synchronized void recover() throws ResourceException {
        if (!isResourcePersistent()) {
            String errorMessage = i18n.getMessage(
                Resources.PRECONDITION_VIOLATION,
                "!this.isResourcePersistent()");
            throw new RuntimeException(errorMessage);
        }

        logger.debug("Loading the home's list of resource keys");

        List resourceKeyValueList = null;
        try {
            XmlPersistenceHelper persistenceHelper = new XmlPersistenceHelper(
                ManagedMultiJobResourceStateType.class);
            resourceKeyValueList = persistenceHelper.list();
        } catch (Exception e) {
            throw new ResourceException(
                i18n.getMessage("persistedResourcePropertiesDirectoryMissing"),
                e);
        }

        Iterator it = resourceKeyValueList.iterator();
        while (it.hasNext()) {
            String keyValue = (String) it.next();
            ResourceKey key = ManagedJobHelper.getResourceKey(keyValue);
            ManagedMultiJobResource resource
                = (ManagedMultiJobResource) find(key);
            //this.add(key, null);
            logger.info("Added resource key " + keyValue +
                        " to recovered home");
        }

        this.recovered = true;
    }

    /**
     * The actual key value is computed automatically by this function
     * as a UUID.
     * @throws Exception
     * @return EndpointReferenceType
     */
    public ResourceKey create(
            Calendar                    initialTerminationTime,
            Calendar                    currentTime,
            String                      localResourceManager,
            JobDescriptionType          jobDescription,
            String                      userSubject,
            String                      localUserId,
            String                      jobID,
            String                      serviceURL)
            throws                      Exception {

        //Do not allow single-job RSLs
        if (   !(jobDescription instanceof MultiJobDescriptionType)
            || (((MultiJobDescriptionType)jobDescription).getJob() == null)) {
            throw new RemoteException(i18n.getMessage(
                "singleJobNotSupported",
                new Object[]{localResourceManager}));
        }

        //cache the notification consumer home while we have a context
        ManagedMultiJobHome.getNotificationConsumerManager();

        SimpleResourceKey key;

        if (jobID != null) {
            key = ManagedJobHelper.getResourceKey(jobID);
        } else {
            key = ManagedJobHelper.getResourceKey(uuidGen.nextUUID());
        }

        //create EPR using the service URL
        EndpointReferenceType epr =
            AddressingUtils.createEndpointReference(serviceURL, key);

        //register with Resolver and get new resolver key
        //ResourceKey resolverKey = ResolverUtils.register(epr);

        //get String alias to EPR
        String handle = ManagedJobHelper.getHandle(epr);

        //add resolver key to EPR
        //epr = ResolverUtils.addResourceHandle(epr, resolverKey);

        //add env var for this endpoint, so it gets propagated
        //to subjobs and applications can make rendezvous calls to Multijob
        //addEPRtoEnvironment(jobDescription, epr);
        RSLHelper.addEnvironmentVariable(
            jobDescription,
            "GLOBUS_GRAM_MULTIJOB_HANDLE",
            //ResolverUtils.getResourceHandle(epr));
            ManagedJobHelper.getHandle(epr));

        ManagedMultiJobResource resource = null;

        try {
            resource = (ManagedMultiJobResource) find(key);
            //resource.setResolverKey(resolverKey);

            //record full handle alias in reosurce so resource doesn't have
            //to decide of the service URL later (alternative: store service
            //URL only, but no real gain and adds code)
            resource.getResourceData().setHandle(handle);
        } catch (org.globus.wsrf.NoSuchResourceException nsre) {
            logger.debug("new resource being created");
        }


        if (resource == null) {

            //create new resource object
            resource = (ManagedMultiJobResource) this.createNewInstance();

            //initialize resource, including resolver key
            ( (ManagedMultiJobResource) resource).initializeWholeState(
                createResourceState(
                    initialTerminationTime,
                    currentTime,
                    (MultiJobDescriptionType) jobDescription,
                    userSubject,
                    localUserId,
                    handle),
                ManagedMultiJobConstants.RESOURCE_PROPERTY_SET,
                key);
            logger.debug(
                "Created ManagedMultiJobResource with key value "
                + key.getValue());
        } //end if resource == null

        //Very important: store the persistent resource initially, because
        //the garbage collecting of the object will not trigger it.
        if (isResourcePersistent()) {
            ((PersistentResource)resource).store();
        }

        this.add(key, resource);

        return key;
    }

    public void remove(ResourceKey key) throws ResourceException {
        ManagedMultiJobResource resource = (ManagedMultiJobResource) find(key);
        logger.debug("acquiring lock on resource");
        synchronized (resource) {
            logger.debug("acquired lock on resource");
            if (logger.isDebugEnabled()) {
                ResourcePropertySet resourceProperties
                    = resource.getResourcePropertySet();
                StateEnumeration state = (StateEnumeration) resourceProperties.
                    get(ManagedJobConstants.RP_STATE).get(0);
                    logger.debug("remove() called\n"
                                + "state: " + state.toString());
            }

            try {
                //ResolverUtils.unregister(resource.getResolverKey());
                //resource.setResolverKey(null);
            }
            catch (Exception e) {
                throw new ResourceException(e);
            }
            logger.debug("releasing lock on resource");
        } //synchronized (resource)

        super.remove(key);
    }

    /**
     * Creates entire initial state of the resource, including internal
     * (non-published) data.
     * @return ManagedMultiJobResourceStateType state object,
     *         serializable to XML by Axis
     */
    private static ManagedMultiJobResourceStateType createResourceState(
        Calendar                            initialTerminationTime,
        Calendar                            currentTime,
        MultiJobDescriptionType             jobDescription,
        String                              userSubject,
        String                              localUserId,
        String                              handle)
    {
        ManagedMultiJobResourceStateType resourceState =
            new ManagedMultiJobResourceStateType();

        //initializing resource properties data:
        ManagedMultiJobResourcePropertiesType resourceBean =
            createResourcePropertiesBean(
                initialTerminationTime,
                currentTime,
                jobDescription,
                userSubject,
                localUserId);
        resourceState.setResourceProperties(resourceBean);

        ManagedMultiJobResourceDataType resourceData =
            new ManagedMultiJobResourceDataType();
        resourceData.setStarted(false);
        int subJobCount
            = ((MultiJobDescriptionType)jobDescription).getJob().length;
        resourceData.setSubJobCount(subJobCount);
        resourceData.setSubscriptionEndpoint(
            new EndpointReferenceType[subJobCount]);
        resourceData.setStageInCount(0);
        resourceData.setPendingCount(0);
        resourceData.setActiveCount(0);
        resourceData.setStageOutCount(0);
        resourceData.setSuspendedCount(0);
        resourceData.setDoneCount(0);
        resourceData.setFailedCount(0);
        resourceData.setHoldingCount(0);
        /*
        resourceData.setResolverKeyString(
            org.globus.exec.service.job.ManagedExecutableJobResource.
            getResolverKeyString(resolverKey));*/
        resourceData.setHandle(handle);

        resourceState.setResourceData(resourceData);

        return resourceState;
    }

    private static
    ManagedMultiJobResourcePropertiesType createResourcePropertiesBean(
        Calendar                            initialTerminationTime,
        Calendar                            currentTime,
        MultiJobDescriptionType                        jobDescription,
        String                              userSubject,
        String                              localUserId)
    {
        ManagedMultiJobResourcePropertiesType jobResourceProperties =
            new ManagedMultiJobResourcePropertiesType();

        //initialize ScheduledResourceTermination port type resource properties:
        jobResourceProperties.setTerminationTime(initialTerminationTime);
        jobResourceProperties.setCurrentTime(currentTime);

        //initialize ManagedJobPortType resource properties:
        jobResourceProperties.setState(StateEnumeration.Unsubmitted);
        jobResourceProperties.setServiceLevelAgreement(
            new ServiceLevelAgreementType(null, jobDescription));//job, multiJob
        jobResourceProperties.setUserSubject(userSubject);
        jobResourceProperties.setLocalUserId(localUserId);

        //initialize ManagedMultiJobPortType resource properties:
        int subJobCount = jobDescription.getJob().length;

        jobResourceProperties.setSubJobEndpoint(
            new EndpointReferenceType[subJobCount]);

        //initialize RendezvousPortType resource properties:
        jobResourceProperties.setCapacity(subJobCount);
        jobResourceProperties.setRegistrantData(null);
        jobResourceProperties.setRendezvousCompleted(false);


        return jobResourceProperties;
    }

    private static UUIDGen uuidGen = UUIDGenFactory.getUUIDGen();

    public static ManagedMultiJobHome getHome()
        throws                              ResourceException
    {
        try {
            InitialContext initialContext = new InitialContext();
            ResourceContext ctx = ResourceContext.getResourceContext();
            String factoryConfigurationLookupString =
                Constants.JNDI_SERVICES_BASE_NAME + ctx.getService() +
                "/configuration";
            if (logger.isDebugEnabled()) {
                logger.debug("JNDI lookup string for service config:" +
                             factoryConfigurationLookupString);
            }
            FactoryServiceConfiguration factoryConfiguration =
                (FactoryServiceConfiguration) initialContext.lookup(
                factoryConfigurationLookupString);

           String jobHomeLookupString = Constants.JNDI_SERVICES_BASE_NAME +
               factoryConfiguration.getManagedMultiJobServicePath() +
               Constants.HOME_NAME;
           if (logger.isDebugEnabled()) {
               logger.debug("JNDI lookup string for ManagedMultiJobHome: "
                           +jobHomeLookupString);
           }
           return (ManagedMultiJobHome) initialContext.lookup(
                jobHomeLookupString);

        } catch(javax.naming.NamingException ne) {
            throw new ResourceException(ne);
        } catch(org.globus.wsrf.ResourceContextException rce) {
            throw new ResourceException(rce);
        }
    }

    private static NotificationConsumerManager NOTIFICATION_CONSUMER_MANAGER
        = null;;

    public static NotificationConsumerManager getNotificationConsumerManager()
    {
        if (NOTIFICATION_CONSUMER_MANAGER == null)
        {
            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String wsTransportProtocol
                = factoryHomeConfiguration.getStagingProtocol();
            if (wsTransportProtocol == null)
            {
                wsTransportProtocol = ServiceHost.getDefaultProtocol();
            }
            if (wsTransportProtocol.equals("http"))
            {
                NOTIFICATION_CONSUMER_MANAGER
                    = NotificationConsumerManager.getInstance();
            } else
            {
                Map properties = new HashMap();
                properties.put(ServiceContainer.CLASS,
                    "org.globus.wsrf.container.GSIServiceContainer");
                NOTIFICATION_CONSUMER_MANAGER
                    = NotificationConsumerManager.getInstance(properties);
            }
        }

        return NOTIFICATION_CONSUMER_MANAGER;
    }

    public boolean isResourcePersistent() {
        return PersistentResource.class.isAssignableFrom(this.resourceClass);
    }

    public int getResourceCount()
    {
        return this.resources.size();
    }

    private static final Log logger =
        LogFactory.getLog(ManagedMultiJobHome.class.getName());

    static I18n i18n = I18n.getI18n(
        org.globus.exec.utils.Resources.class.getName());

    private static PerformanceLog performanceLogger = new PerformanceLog(
        ManagedMultiJobHome.class.getName() + ".performance");
}
