/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.multi;

import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.security.auth.Subject;

import javax.xml.namespace.QName;
import javax.xml.rpc.Stub;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.addressing.ReferencePropertiesType;

import org.ietf.jgss.GSSCredential;
import org.oasis.wsn.SubscriptionManager;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsn.WSBaseNotificationServiceAddressingLocator;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsrf.lifetime.Destroy;
import org.w3c.dom.Element;

import org.globus.axis.gsi.GSIConstants;
import org.globus.axis.util.Util;
import org.globus.exec.generated.CreateManagedJobInputType;
import org.globus.exec.generated.CreateManagedJobOutputType;
import org.globus.exec.generated.FaultResourcePropertyType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedJobFactoryPortType;
import org.globus.exec.generated.ManagedJobPortType;
import org.globus.exec.generated.ManagedMultiJobResourceDataType;
import org.globus.exec.generated.ManagedMultiJobResourceStateType;
import org.globus.exec.generated.MultiJobDescriptionType;
import org.globus.exec.generated.NameValuePairType;
import org.globus.exec.generated.ReleaseInputType;
import org.globus.exec.generated.StateChangeNotificationMessageType;
import org.globus.exec.generated.StateChangeNotificationMessageWrapperType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.service.job.ManagedJobResourceImpl;
import org.globus.exec.service.utils.DelegatedCredential;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.utils.JobDescriptionMap;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.ManagedMultiJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.utils.client.ManagedJobClientHelper;
import org.globus.exec.utils.client.ManagedJobFactoryClientHelper;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.security.gridmap.GridMap;
import org.globus.util.I18n;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.Topic;
import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.container.ServiceContainer;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.SimpleTopic;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.Authorization;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.GSISecureMsgAuthMethod;
import org.globus.wsrf.impl.security.descriptor.GSITransportAuthMethod;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.security.SecureResource;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.XmlUtils;

public class ManagedMultiJobResource
    extends                                 ManagedJobResourceImpl
    implements                              RemoveCallback,
                                            SecureResource,
                                            NotifyCallback,
                                            Runnable
{

    public void initializeWholeState(
        Object                              wholeResourceState,
        QName                               resourceElementQName,
        Object                              key)
        throws                              ResourceException
    {
        super.initializeWholeState(
            wholeResourceState, resourceElementQName, key);

        ManagedMultiJobResourceStateType wholeJobState =
            (ManagedMultiJobResourceStateType)wholeResourceState;
        setResourceData(wholeJobState.getResourceData());
        this.initialize(wholeJobState.getResourceProperties(),
                        resourceElementQName, key);
    }

    public void initialize(
        Object                              resourceBean,
        QName                               resourceElementQName,
        Object                              key)
        throws                              ResourceException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering initialize()");
        }

        super.initialize(resourceBean, resourceElementQName, key); //calls
        //initializeTopics()

        //Create the job data cache
        String localUserId
            = (String) this.getResourcePropertySet().get(
                ManagedJobConstants.RP_LOCAL_USER_ID).get(0);

        initSecurity();

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving initialize()");
        }
    }

    public void initSecurity() throws ResourceException {
        //get security config
        ResourceSecurityConfig securityConfig= new ResourceSecurityConfig(
            ManagedJobConstants.SECURITY_CONFIG_FILE_NAME);
        try {
            securityConfig.init();
        } catch (Exception e) {
            throw new ResourceException(
                i18n.getMessage("resourceSecurityInitializationFailed"),
                e);
        }

        //get security descriptor
        this.securityDescriptor = securityConfig.getSecurityDescriptor();

        //Set the gridmap from resource properties set during resource creation
        String userSubject
            = (String) this.getResourcePropertySet().get(
                ManagedJobConstants.RP_USER_SUBJECT).get(0);
        String localUserId
            = (String) this.getResourcePropertySet().get(
                ManagedJobConstants.RP_LOCAL_USER_ID).get(0);
        if (logger.isDebugEnabled()) {
            logger.debug("setting resource securty grid map..."
                        + "\nSubject: " + userSubject
                        + "\nLocal user ID: " + localUserId);
        }
        GridMap gridMap = new GridMap();
        gridMap.map(userSubject, localUserId);
        this.securityDescriptor.setGridMap(gridMap);
    }

    protected void initializeTopics() {
        super.initializeTopics();
        Topic stateChangeTopic = new SimpleTopic(ManagedJobConstants.RP_STATE);
        this.getTopicList().addTopic(stateChangeTopic);
        //requires manual notifications in relevant parts of the code
    }

    public void remove()
        throws                              ResourceException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering remove()");
        }

        if ((getResourceData() != null) && getResourceData().isStarted()) {
            destroySubJobs();
        }

        logger.debug(
            "deleting cache directory for job resource " +
            getID());

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving remove()");
        }
    }

    /* Start the state machine.
     */
    public synchronized void start()
    {
        boolean started
            = ((Boolean)getResourceDatum(ManagedJobConstants.RD_STARTED))
            .booleanValue();
        if (!started)
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Entering start()");
            }

            //Start the state machine
            StateEnumeration state
                = (StateEnumeration) getResourcePropertySet()
                .get(ManagedJobConstants.RP_STATE).get(0);

            if (state.equals(StateEnumeration.Failed)) {
                return;
            }

            new Thread(this).start();
        }
    }

    public void run()
    {
        FaultType fault = null;
        DelegatedCredential jobCredential = null;
        try
        {
            jobCredential = getJobCredential(null);
        }
        catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                Resources.CREDENTIAL_ACQUISITION_ERROR);
            fault = FaultUtils.makeFault(
                FaultType.class,
                (ResourceKey) getID(),
                errorMessage,
                e,
                getState(),
                null,
                0);
        }
        if ((jobCredential == null) && (fault == null))
        {
            String faultMessage
                = i18n.getMessage(Resources.MISSING_JOB_CREDENTIAL);
            fault = FaultUtils.makeFault(
                FaultType.class,
                (ResourceKey) getID(),
                faultMessage,
                null,
                (StateEnumeration) getResourceProperty(
                    ManagedJobConstants.RP_STATE),
                null,
                0);
        }

        if (fault == null)
        {
            try
            {
                createSubJobs();
            }
            catch (Exception e)
            {
                String faultMessage = i18n.getMessage("subJobsCreationFailed");
                fault = FaultUtils.makeFault(
                    FaultType.class,
                    (ResourceKey) getID(),
                    faultMessage,
                    e,
                    (StateEnumeration) getResourceProperty(
                        ManagedJobConstants.RP_STATE),
                    null,
                    0);
            }
        }

        if (fault != null)
        {
            setFault(fault);

            try
            {
                setState(StateEnumeration.Failed);
            }
            catch (Exception e)
            {
                String errorMessage = i18n.getMessage(
                    Resources.MULTI_RESOURCE_SET_STATE_FAILED_NOTIFY_ERROR);
                throw new RuntimeException(errorMessage, e);
            }
        }
        else
        {
            getResourceData().setStarted(true);
        }
    }

    private void createSubJobs()
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering createSubJobs()");
        }

        JobDescriptionType jobDescription = getJobDescription();

        ResourceProperty subJobEndpointsRP
            = this.getResourcePropertySet().get(
                ManagedMultiJobConstants.RP_SUB_JOB_ENDPOINTS);

        //get multi-job description
        MultiJobDescriptionType multiJobDescription
            = (MultiJobDescriptionType) jobDescription;

        //breakup mulit-job RSL into separate RSLs for each sub-job
        JobDescriptionType[] subJobDescriptions
            = ManagedMultiJobResource.breakUpMultiJobDescription(
                multiJobDescription,
                getResourceData().getHandle());

        //call createSubJob() for each sub-job RSL.
        for (int index=0; index<subJobDescriptions.length; index++) {
            if (logger.isDebugEnabled()) {
                logger.debug("Sub-Job RSL:\n"
                    + RSLHelper.convertToString(subJobDescriptions[index]));
            }
            EndpointReferenceType subJobEndpoint
                = createSubJob(subJobDescriptions[index], index);

            subJobEndpointsRP.set(index, subJobEndpoint);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving createSubJobs()");
        }
    }

    private static JobDescriptionType[] breakUpMultiJobDescription(
        MultiJobDescriptionType             multiJobDescription,
        String                              multiJobHandle)
        throws                              ResourceException
    {
        //get sub-job descriptions
        JobDescriptionType[] subJobDescriptions = multiJobDescription.getJob();

        //propagate environment variables defined in the multijob
        //(the default Reflection-based propagation would not do it if merely
        //one env var is defined in a subjob)
        NameValuePairType[] multiEnvVars = multiJobDescription.getEnvironment();
        if (multiEnvVars != null) {
            for (int i = 0; i < multiEnvVars.length; i++) {
                for (int j = 0; j < subJobDescriptions.length; j++) {
                    RSLHelper.addEnvironmentVariable(subJobDescriptions[j],
                        multiEnvVars[i].getName(),
                        multiEnvVars[i].getValue());
                }
            }
        }

        //add respective ranks to subjobs as an env variable
        for (int i = 0; i < subJobDescriptions.length; i++) {
            RSLHelper.addEnvironmentVariable(
                subJobDescriptions[i],
                "GLOBUS_GRAM_SUBJOB_RANK",
                Integer.toString(i));
        }

        //explicitly set null sub-job attributes based on multi-job defaults
        JobDescriptionMap multiJobAttributes = null;

        //get attribute map for multi-job description
        try {
            multiJobAttributes = new JobDescriptionMap(multiJobDescription);
        } catch (Exception e) {
            throw new ResourceException(
                i18n.getMessage(
                    "objectCreationFailed",
                    new Object[]{JobDescriptionMap.class.getName()}),
            e);

        }

        String[] attributeSkipList = new String[]
        {
            "job",
            "jobCredentialEndpoint",
            "stagingCredentialEndpoint",
            "factoryEndpoint"
        };

        //use muli-job attributes as defaults
        //to set null sub-job attributes
        Iterator attributeNames = multiJobAttributes.keySet().iterator();
        while (attributeNames.hasNext()) {
            String attributeName = (String) attributeNames.next();

            boolean skip = false;
            for (int index=0; index<attributeSkipList.length; index++)
            {
                if (attributeName.equals(attributeSkipList[index]))
                {
                    skip = true;
                }
            }
            if (skip)
            {
                continue;
            }

            PropertyDescriptor propertyDescriptor = null;
            try {
                propertyDescriptor = new PropertyDescriptor(
                        attributeName, JobDescriptionType.class);
            } catch (Exception e) {
                throw new ResourceException(
                    i18n.getMessage(
                        "objectCreationFailed",
                        new Object[]{PropertyDescriptor.class.getName()}),
                e);

            }
            Method readMethod = propertyDescriptor.getReadMethod();
            Method writeMethod = propertyDescriptor.getWriteMethod();

            boolean allMustHold = false;
            StateEnumeration holdState = null;
            if (attributeName.equals("holdState")) {
                allMustHold = true;
                holdState
                    = (StateEnumeration) multiJobAttributes.get(attributeName);
            }

            for (int index=0; index<subJobDescriptions.length; index++) {
                Object oldValue = null;
                try {
                    oldValue = readMethod.invoke(
                        subJobDescriptions[index], null);
                } catch (Exception e) {
                    throw new ResourceException(
                        i18n.getMessage(
                            "beanMethodInvocationFailed",
                            new Object[]{
                                subJobDescriptions[index].getClass().getName(),
                                readMethod.getName()}),
                    e);

                }

                //use multi-job default to set null sub-job attribute
                if (oldValue == null) {
                    try {
                        writeMethod.invoke(
                            subJobDescriptions[index],
                            new Object[]{
                                multiJobAttributes.get(attributeName)});
                    } catch (Exception e) {
                        throw new ResourceException(
                            i18n.getMessage(
                                "beanMethodInvocationFailed",
                                new Object[]{
                                    subJobDescriptions[index].getClass().
                                        getName(),
                                    writeMethod.getName()}),
                        e);

                    }
                } else {
                    //check for disagreeing holdState values
                    if (attributeName.equals("holdState")) {
                        if (allMustHold && !holdState.equals(oldValue)) {
                            //this sub-job's holdState disagrees wither with
                            //the multi-job's or another sub-job's holdState
                            throw new ResourceException(
                                i18n.getMessage(
                                    "unexpectedHoldState",
                                    new Object[]{ holdState, oldValue }
                                )
                            );
                        } else if (!allMustHold && (index > 0)) {
                            //both the multi-job and preceding sub-jobs
                            //do not have holdState values
                            throw new ResourceException(
                                i18n.getMessage(
                                    "unexpectedHoldState",
                                    new Object[]{ "null", oldValue }
                                )
                            );
                        } else {
                            //no multi-job holdState, but this is the first
                            //sub-job and it has a holdState. use for consensus.
                            allMustHold = true;
                            holdState = (StateEnumeration) oldValue;
                        }
                    }
                }
            }
        }

        return subJobDescriptions;
    }

    private EndpointReferenceType createSubJob(
        JobDescriptionType                  subJobDescription,
        int                                 subJobIndex)
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering createSubJob()");
        }

        //get sub-job factory endpoint
        EndpointReferenceType factoryEndpoint =
            subJobDescription.getFactoryEndpoint();
        if (factoryEndpoint == null) {
            throw new ResourceException(
                i18n.getMessage("jobDescriptionMissingEndpoint"));
        }

        //setup MJFS stub based on endpoint in subJobDescription
        ManagedJobFactoryPortType factoryPort = null;
        try {
            factoryPort
                = ManagedJobFactoryClientHelper.getPort(factoryEndpoint);
        } catch (Exception e) {
            throw new ResourceException(e);
        }

        setStubSecurity((Stub) factoryPort);

        //setup the resource to listen for sub-job notifications,
        EndpointReferenceType notificationConsumerEndpoint = null;
        try {
            NotificationConsumerManager notificationConsumerManager
                = ManagedMultiJobHome.getNotificationConsumerManager();
            List topicPath = new LinkedList();
            topicPath.add(ManagedJobConstants.RP_STATE);
            ResourceSecurityDescriptor securityDescriptor
                = new ResourceSecurityDescriptor();
            securityDescriptor.setAuthz(Authorization.AUTHZ_NONE);
            Vector authMethod = new Vector();

            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            String wsTransportProtocol
                = factoryHomeConfiguration.getStagingProtocol();
            if (wsTransportProtocol == null)
            {
                wsTransportProtocol = ServiceHost.getDefaultProtocol();
            }
            if (wsTransportProtocol.equals("http"))
            {
                authMethod.add(GSISecureMsgAuthMethod.BOTH);
            } else
            {
                authMethod.add(GSITransportAuthMethod.BOTH);
            }
            securityDescriptor.setAuthMethods(authMethod);
            notificationConsumerManager.startListening();

            notificationConsumerEndpoint =
                notificationConsumerManager.createNotificationConsumer(
                    topicPath,
                    this,
                    securityDescriptor);

            if (logger.isDebugEnabled()) {
                logger.debug("notification consumer endpoint: "
                            + notificationConsumerEndpoint);
            }
        } catch (Exception e) {
                String errorMessage = i18n.getMessage(
                Resources.NOTIFICATION_CONSUMER_CREATION_ERROR);
            throw new RuntimeException(errorMessage, e);
        }
        setResourceDatum(
            ManagedJobConstants.RD_NOTIFICATION_CONSUMER_ENDPOINT,
            notificationConsumerEndpoint);

        //setup subscription message
        Subscribe subscribeRequest = new Subscribe();
        subscribeRequest .setConsumerReference(notificationConsumerEndpoint);
        TopicExpressionType topicExpression = null;
        try {
            topicExpression = new TopicExpressionType(
                WSNConstants.SIMPLE_TOPIC_DIALECT,
                ManagedJobConstants.RP_STATE);
        } catch (Exception e) {
            throw new RuntimeException(
                i18n.getMessage(
                    "objectCreationFailed",
                    new Object[]{TopicExpressionType.class.getName()}),
            e);

        }
        subscribeRequest .setTopicExpression(topicExpression);

        //call createManagedJob(subJobDescription) on MJFS stub
        CreateManagedJobInputType jobCreationInput =
            new CreateManagedJobInputType();
        //jobInput.setInitialTerminationTime();
        jobCreationInput.setJob(subJobDescription);
        jobCreationInput.setSubscribe(subscribeRequest );
        CreateManagedJobOutputType response
            = factoryPort.createManagedJob(jobCreationInput);

        //save the subscription endpoint
        ManagedMultiJobResourceDataType resourceData
            = (ManagedMultiJobResourceDataType) getResourceData();
        EndpointReferenceType subscriptionEndpoint
            = response.getSubscriptionEndpoint();
        resourceData.getSubscriptionEndpoint()[subJobIndex]
            = response.getSubscriptionEndpoint();

        //get the job endpoint
        EndpointReferenceType jobEndpoint = response.getManagedJobEndpoint();

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving createSubJob()");
        }

        return jobEndpoint;
    }

    private static String getResourceIdFromEndpoint(
        EndpointReferenceType               endpoint)
    {
        //get the resource ID from the endpoint
        if (endpoint == null)
        {
            return "null";
        }
        ReferencePropertiesType subJobEprReferenceProperties
            = endpoint.getProperties();
        MessageElement[] any = subJobEprReferenceProperties.get_any();
        Element element = null;
        try {
            element = AnyHelper.toElement(any[0]);
        } catch (Exception ee) {
            String errorMessage =
                i18n.getMessage(Resources.ANY_TO_ELEMENT_CONVERSION_ERROR);
            logger.error(errorMessage, ee);
        }
        logger.debug("Reference Properties: "
                    +XmlUtils.toString(element));
        String resourceId = null;
        if (element.hasChildNodes()) {
            resourceId = element.getFirstChild().getNodeValue();
        }
        logger.debug("Resource ID: " + resourceId);

        return resourceId;
    }

    void releaseSubJobs()
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering releaseSubJobs()");
        }

        JobDescriptionType jobDescription = getJobDescription();

        ResourceProperty subJobEndpointsRP
            = this.getResourcePropertySet().get(
                ManagedMultiJobConstants.RP_SUB_JOB_ENDPOINTS);

        //get multi-job description
        MultiJobDescriptionType multiJobDescription
            = (MultiJobDescriptionType) jobDescription;

        //breakup mulit-job JobDescription into separate JobDescriptions for
        //each sub-job
        JobDescriptionType[] subJobDescriptions
            = ManagedMultiJobResource.breakUpMultiJobDescription(
                multiJobDescription,
                getResourceData().getHandle());

        int subJobCount = ((Integer) getResourceDatum(
            ManagedMultiJobConstants.RD_SUB_JOB_COUNT)).intValue();
        for (int index=0; index<subJobCount; index++)
        {
            EndpointReferenceType subJobEndpoint
                = (EndpointReferenceType) subJobEndpointsRP.get(index);

            releaseSubJob(subJobEndpoint, subJobDescriptions[index]);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving releaseSubJobs()");
        }
    }

    private void releaseSubJob(
        EndpointReferenceType               subJobEndpoint,
        JobDescriptionType                  subJobDescription)
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering releaseSubJob()");
        }

        //setup MJS stub
        ManagedJobPortType subJobPort = null;
        try {
            subJobPort
                = ManagedJobClientHelper.getPort(subJobEndpoint);
        } catch (Exception e) {
            throw new ResourceException(e);
        }

        setStubSecurity((Stub) subJobPort);

        subJobPort.release(new ReleaseInputType());

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving releaseSubJob()");
        }
    }

    private void destroySubJobs()
        throws                              ResourceException
    {

        if (logger.isDebugEnabled()) {
            logger.debug("Entering destroySubJobs()");
        }

        ManagedMultiJobResourceDataType resourceData
            = (ManagedMultiJobResourceDataType) getResourceData();

        logger.trace("Obtaining lock on resourceData");
        synchronized (resourceData)
        {
            logger.trace("Obtained lock on resourceData");

            int subJobCount = ((Integer) getResourceDatum(
                ManagedMultiJobConstants.RD_SUB_JOB_COUNT)).intValue();

            String errorMessage = null;
            Exception exception = null;

            //prevent multiple calls to destroySubJobs();
            if (!resourceData.isDestroyCalled())
            {
                try {
                    unsubscribeToNotificationsForSubJobs();
                } catch (Exception e) {
                    errorMessage = i18n.getMessage(
                        Resources.SUBJOB_UNSUBSCRIBE_ERROR);
                    exception = e;

                    logger.debug(errorMessage, exception);
                }

                JobDescriptionType jobDescription = getJobDescription();

                //get multi-job description
                MultiJobDescriptionType multiJobDescription
                    = (MultiJobDescriptionType) jobDescription;

                //breakup mulit-job JobDescription into separate JobDescriptions
                //for each sub-job
                JobDescriptionType[] subJobDescriptions
                    = ManagedMultiJobResource.breakUpMultiJobDescription(
                        multiJobDescription,
                        getResourceData().getHandle());

                //call destroy() on all of the sub-jobs
                ResourceProperty subJobEndpointsRP
                    = this.getResourcePropertySet().get(
                        ManagedMultiJobConstants.RP_SUB_JOB_ENDPOINTS);
                for (int index=0; index<subJobCount; index++) {
                    EndpointReferenceType subJobEndpoint
                        = (EndpointReferenceType) subJobEndpointsRP.get(index);

                    try {
                        destroySubJob(
                            subJobEndpoint,
                            subJobDescriptions[index]);
                    } catch (Exception e) {
                        /*Don't prevent the multi-job resource from being
                         *destroyed even if one or more sub-job resoruces
                         *can't be destroyed*/
                        errorMessage
                            = i18n.getMessage(
                                "subJobDestroyFailed",
                                new Object[] {
                                    ManagedMultiJobResource.
                                        getResourceIdFromEndpoint(
                                            subJobEndpoint) });
                        exception = e;

                        logger.debug(errorMessage, exception);
                    }
                }
            } else {
                logger.debug("Ignoring second call to destroySubJobs()");
            }


            ((ManagedMultiJobResourceDataType)getResourceData())
                .setDestroyCalled(true);

            if (exception != null)
            {
                throw new ResourceException(errorMessage, exception);
            }

            logger.trace("Releasing lock on resourceData");
        }
        logger.trace("Released lock on resourceData");

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving destroySubJobs()");
        }
    }

    private void unsubscribeToNotificationsForSubJobs()
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering unsubscribeToNotificationsForSubJobs()");
        }

        //destroy notification consumer
        EndpointReferenceType notificationConsumerEndpoint
            = getResourceData().getNotificationConsumerEndpoint();
        if (notificationConsumerEndpoint != null)
        {
            try
            {
                NotificationConsumerManager notificationConsumerManager = null;
                FactoryHomeConfiguration factoryHomeConfiguration
                    = ManagedJobFactoryHome.getConfiguration();
                String wsTransportProtocol
                    = factoryHomeConfiguration.getStagingProtocol();
                if (wsTransportProtocol == null)
                {
                    wsTransportProtocol = ServiceHost.getDefaultProtocol();
                }
                if (wsTransportProtocol.equals("http"))
                {
                    notificationConsumerManager =
                        NotificationConsumerManager.getInstance();
                } else
                {
                    Map properties = new HashMap();
                    properties.put(ServiceContainer.CLASS,
                        "org.globus.wsrf.container.GSIServiceContainer");
                    notificationConsumerManager =
                        NotificationConsumerManager.getInstance(properties);
                }
                notificationConsumerManager.startListening();
                notificationConsumerManager.removeNotificationConsumer(
                    notificationConsumerEndpoint);
            }
            catch (Exception e)
            {
                if (!getState().equals(StateEnumeration.Failed))
                {
                String errorMessage = i18n.getMessage(
                    Resources.NOTIFICATION_CONSUMER_REMOVAL_ERROR);

                    logger.debug(errorMessage, e);

                    throw new RemoteException(errorMessage, e);
                }

                return;
            }
        }

        JobDescriptionType jobDescription = getJobDescription();

        //get multi-job description
        MultiJobDescriptionType multiJobDescription
            = (MultiJobDescriptionType) jobDescription;

        //breakup mulit-job JobDescription into separate JobDescriptions for
        //each sub-job
        JobDescriptionType[] subJobDescriptions
            = ManagedMultiJobResource.breakUpMultiJobDescription(
                multiJobDescription,
                getResourceData().getHandle());

        //unsubscribe to notifications for sub-job states
        int subJobCount = ((Integer) getResourceDatum(
            ManagedMultiJobConstants.RD_SUB_JOB_COUNT)).intValue();
        for (int index=0; index<subJobCount; index++) {
            EndpointReferenceType subscriptionEndpoint = null;
            try {
                subscriptionEndpoint
                    = ((ManagedMultiJobResourceDataType)getResourceData())
                        .getSubscriptionEndpoint(index);
                if (subscriptionEndpoint != null) {
                    SubscriptionManager subscriptionPort
                        = new WSBaseNotificationServiceAddressingLocator().
                            getSubscriptionManagerPort(subscriptionEndpoint);

                    try
                    {
                        setStubSecurity((Stub) subscriptionPort);
                    }
                    catch (Exception e)
                    {
                        if (!getState().equals(StateEnumeration.Failed))
                        {
                            String errorMessage = i18n.getMessage(
                                Resources.SUBJOB_STUB_SECURITY_ERROR);

                            logger.debug(errorMessage, e);

                            throw new RemoteException(errorMessage, e);
                        }

                        return;
                    }

                    subscriptionPort.destroy(new Destroy());
                }
            } catch (Exception e) {
                EndpointReferenceType subJobEndpoint = null;
                ResourceProperty subJobEndpointsRP
                    = this.getResourcePropertySet().get(
                        ManagedMultiJobConstants.RP_SUB_JOB_ENDPOINTS);
                subJobEndpoint = (EndpointReferenceType)
                    subJobEndpointsRP.get(index);
                throw new RemoteException(
                    i18n.getMessage(
                            "subJobUnsubscribeFailed",
                            new Object[] {
                                ManagedMultiJobResource.
                                    getResourceIdFromEndpoint(subJobEndpoint) }),
                    e);
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving unsubscribeToNotificationsForSubJobs()");
        }
    }

    private void destroySubJob(
        EndpointReferenceType               subJobEndpoint,
        JobDescriptionType                  subJobDescription)
        throws                              RemoteException
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering destroySubJob()");
        }

        if (subJobEndpoint == null)
        {
            logger.debug(
                "Leaving destroySubJob() since subJobEndpoint is null");
            return;
        }

        //setup MJS stub
        ManagedJobPortType subJobPort = null;
        try {
            subJobPort
                = ManagedJobClientHelper.getPort(subJobEndpoint);
        } catch (Exception e) {
            throw new ResourceException(e);
        }

        try
        {   
            setStubSecurity((Stub) subJobPort);
        }
        catch (RemoteException re)
        {
            if (getState().equals(StateEnumeration.Failed))
            {
                String faultDescription
                    = getFault().getDescription().toString();
                logger.debug("Fault Description: " + faultDescription);
                if (faultDescription.equals(
                        i18n.getMessage("subJobsCreationFailed")))
                {
                    return;
                }
                else
                {
                    throw re;
                }
            }
            else
            {
                throw re;
            }
        }

        subJobPort.destroy(new Destroy());

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving destroySubJob()");
        }
    }

    private void setStubSecurity(
        Stub                                stub)
        throws                              RemoteException
    {
        //determine security type (message-level or transport-level)
        FactoryHomeConfiguration factoryHomeConfiguration
            = ManagedJobFactoryHome.getConfiguration();
        String wsTransportProtocol
            = factoryHomeConfiguration.getStagingProtocol();
        String securityType = Constants.GSI_TRANSPORT;
        if (wsTransportProtocol == null)
        {
            wsTransportProtocol = ServiceHost.getDefaultProtocol();
        }
        if (wsTransportProtocol.equals("http"))
        {
            securityType = Constants.GSI_SEC_MSG;
        }

        ClientSecurityDescriptor secDesc = new ClientSecurityDescriptor();

        //set security type
        if (securityType.equals(Constants.GSI_SEC_MSG))
        {
            secDesc.setGSISecureMsg(Constants.SIGNATURE);
        }
        else
        {
            secDesc.setGSITransport(Constants.SIGNATURE);
        }

        try {
            secDesc.setGSSCredential(
                new GlobusGSSCredentialImpl(
                    getJobCredential(null).getCredential(),
                    GSSCredential.INITIATE_AND_ACCEPT));
        } catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.GLOBUS_CREDENTIAL_TO_GSS_CONVERSION_ERROR);
            throw new RemoteException(errorMessage, e);
        }

        ContainerSecurityConfig containerSecurityConfig = null;
        ContainerSecurityDescriptor containerSecurityDescriptor = null;
        Subject containerSubject = null;
        Set principals = null;

        try
        {
            containerSecurityConfig
                = ContainerSecurityConfig.getConfig();
            containerSecurityDescriptor
                = containerSecurityConfig.getSecurityDescriptor();
            containerSubject
                = containerSecurityDescriptor.getSubject();
            if (containerSubject != null)
            {
                principals = containerSubject.getPrincipals();
            }
        } catch (Exception e)
        {
            String errorMessage = i18n.getMessage(
                Resources.CREDENTIAL_CONFIGURATION_ERROR);
            throw new RuntimeException(errorMessage);
        }

        if (containerSubject != null)
        {

            if (containerSecurityConfig.getSecurityDescriptorFile() == null)
            {
                //using default creds, assume identity authz
                if (logger.isDebugEnabled()) {
                    logger.debug("Using default creds. "
                                +"Assume identity authz");
                }
                secDesc.setAuthz(new IdentityAuthorization(
                    principals.iterator().next().toString()));
            }
            else
            if (containerSecurityDescriptor.getProxyFilename() != null)
            {
                logger.debug("Found proxy file configuration");

                secDesc.setAuthz(new IdentityAuthorization(
                    principals.iterator().next().toString()));
            }
            else
            if (   (containerSecurityDescriptor.getCertFilename() != null)
                && (containerSecurityDescriptor.getKeyFilename() != null))
            {
                logger.debug("Found host and cert configuration.");

                secDesc.setAuthz(HostAuthorization.getInstance());
            }
            else
            {
                logger.debug("No credentials specified, assuming user creds");

                secDesc.setAuthz(new IdentityAuthorization(
                    principals.iterator().next().toString()));
            }
        } else {
            //no container credentials

            String errorMessage = i18n.getMessage(
                Resources.CREDENTIAL_CONFIGURATION_ERROR);
            throw new RuntimeException(errorMessage);
        }

        stub._setProperty(Constants.CLIENT_DESCRIPTOR,
                          secDesc);
    }

    public void setState(
        StateEnumeration                    state)
        throws                              Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug("receiving request for state change to " +
                         state.getValue());
        }

        super.setState(state);

        //set the new state value
        if (logger.isDebugEnabled()) {
            logger.debug("State of job " + this.getID()
                        + " changed to: " + getState().toString());
        }

        //fetch the holding flag
        Boolean holding = (Boolean) getResourcePropertySet().get(
            ManagedJobConstants.RP_HOLDING).get(0);
        if (logger.isDebugEnabled()) {
            logger.debug("Holding: " + holding);
        }

        //fetch the fault if the state is 'Failed'
        ResourceProperty faultRP
            = this.getResourcePropertySet().get(ManagedJobConstants.RP_FAULT);
        FaultResourcePropertyType fault = null;
        if (state.equals(StateEnumeration.Failed)){
            fault = (FaultResourcePropertyType) faultRP.get(0);
        }

        //create notification message
        StateChangeNotificationMessageType message
            = new StateChangeNotificationMessageType();
        message.setState(state);
        message.setHolding(holding.booleanValue());
        message.setFault(fault);

        StateChangeNotificationMessageWrapperType messageWrapper
            = new StateChangeNotificationMessageWrapperType();
        messageWrapper.setStateChangeNotificationMessage(message);

        //notify on the topic using the state change message
        LinkedList topicPath = new LinkedList();
        topicPath.add(ManagedJobConstants.RP_STATE);
        Topic stateChangeTopic = this.getTopicList().getTopic(topicPath);
        logger.debug("Notifying of job state change to topic listeners");

        stateChangeTopic.notify(messageWrapper);
    }

    //=== NotificationConsumerCallbackManager methods ===//
    public void registerCallback(
        List                                topicPath,
        NotifyCallback                      callback)
    {
        //NO-OP
    }

    public NotifyCallback getCallback(
        List                                topicPath)
    {
        return this;
    }

    public void deliver(
        List                                topicPath,
        EndpointReferenceType               producerEndpoint,
        Object                              messageWrapper)
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering deliver()");
            logger.debug("Notification Message Type: "
                        +messageWrapper.getClass().getName());
        }

        try {
            StateChangeNotificationMessageType message
                = ((StateChangeNotificationMessageWrapperType) messageWrapper).
                    getStateChangeNotificationMessage();
            StateEnumeration state = message.getState();
            boolean holding = message.isHolding();
            FaultResourcePropertyType fault = message.getFault();

            processStateNotification(state, holding, fault);
        }
        catch (Exception e) {
            String errorMessage = i18n.getMessage(
                Resources.STATE_NOTIFICATION_PROCESSING_ERROR);
            logger.error(errorMessage, e);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving deliver()");
        }
    }

    public void processStateNotification(
        StateEnumeration                    state,
        boolean                             holding,
        FaultResourcePropertyType           faultWrapper)
        throws                              Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Entering processStateNotification()");
            logger.debug("Processing State " + state.toString());
        }

        ManagedMultiJobResourceDataType resourceData
            = (ManagedMultiJobResourceDataType) getResourceData();

        int subJobCount = ((Integer) getResourceDatum(
            ManagedMultiJobConstants.RD_SUB_JOB_COUNT)).intValue();

        StateEnumeration currentState = (StateEnumeration)
            getResourceProperty(ManagedJobConstants.RP_STATE);

        int holdingCount = ((Integer) getResourceDatum(
            ManagedMultiJobConstants.RD_HOLDING_COUNT)).intValue();
        boolean allHolding = ((Boolean) getResourceProperty(
            ManagedJobConstants.RP_HOLDING)).booleanValue();
        if (holding) {
            //increment sub-job holding flag count
            resourceData.setHoldingCount(++holdingCount);

            //set global holding flag and notify of holding state
            //when all sub-jobs are holding
            if (holdingCount == subJobCount) {
                setResourceProperty(
                    ManagedJobConstants.RP_HOLDING,
                    Boolean.TRUE);
                setState(state);
            }

        }
        else
        {
            if ((holdingCount > 0) && allHolding) {
                //turn off global holding flag since at least one sub-job
                //has been released
                setResourceProperty(
                    ManagedJobConstants.RP_HOLDING,
                    Boolean.FALSE);

                //reset sub-job holding count
                resourceData.setHoldingCount(0);
            }

            if (state.equals(StateEnumeration.StageIn)
                && !currentState.equals(StateEnumeration.Pending)
                && !currentState.equals(StateEnumeration.Active)
                && !currentState.equals(StateEnumeration.StageOut)
                && !currentState.equals(StateEnumeration.CleanUp)
                && !currentState.equals(StateEnumeration.Done)
                && !currentState.equals(StateEnumeration.Failed)) {
                int stageInCount = resourceData.getStageInCount();
                resourceData.setStageInCount(++stageInCount);
                if (stageInCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.Pending)
                && !currentState.equals(StateEnumeration.Active)
                && !currentState.equals(StateEnumeration.StageOut)
                && !currentState.equals(StateEnumeration.CleanUp)
                && !currentState.equals(StateEnumeration.Done)
                && !currentState.equals(StateEnumeration.Failed)) {
                int pendingCount = resourceData.getPendingCount();
                resourceData.setPendingCount(++pendingCount);
                if (pendingCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.Active)
                && !currentState.equals(StateEnumeration.StageOut)
                && !currentState.equals(StateEnumeration.CleanUp)
                && !currentState.equals(StateEnumeration.Done)
                && !currentState.equals(StateEnumeration.Failed)) {
                int activeCount = resourceData.getActiveCount();
                resourceData.setActiveCount(++activeCount);
                if (logger.isDebugEnabled()) {
                    logger.debug("Setting activeCount to " + activeCount);
                }
                if (activeCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.StageOut)
                && !currentState.equals(StateEnumeration.CleanUp)
                && !currentState.equals(StateEnumeration.Done)
                && !currentState.equals(StateEnumeration.Failed)) {
                int stageOutCount = resourceData.getStageOutCount() + 1;
                resourceData.setStageOutCount(stageOutCount);
                if (stageOutCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.CleanUp)
                && !currentState.equals(StateEnumeration.Done)
                && !currentState.equals(StateEnumeration.Failed)) {
                int cleanUpCount = resourceData.getCleanUpCount() + 1;
                resourceData.setCleanUpCount(cleanUpCount);
                if (cleanUpCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.Suspended)) {
                int suspendedCount = resourceData.getSuspendedCount();
                resourceData.setSuspendedCount(suspendedCount);
                if (suspendedCount == subJobCount) {
                    setState(state);
                }
            } else
            if (state.equals(StateEnumeration.Done)) {
                int doneCount = resourceData.getDoneCount() + 1;
                resourceData.setDoneCount(doneCount);
                if (logger.isDebugEnabled()) {
                    logger.debug("Setting doneCount to " + doneCount);
                }
                if (doneCount == subJobCount) {
                    setState(state);
                }
            }
            else if (state.equals(StateEnumeration.Failed))
            {
                logger.trace("setting fault");
                setResourceProperty(
                    ManagedJobConstants.RP_FAULT,
                    faultWrapper);

                logger.trace("setting Failed state");
                try {
                    setState(StateEnumeration.Failed);
                } catch (Exception e) {
                    throw new ResourceException(" ", e);
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Leaving processStateNotification()");
        }
    }

    /**
     * Timestamp of last stage change event. This may need to be persistent.
     *
    private Date lastStateTimestamp = new java.util.Date();*/

    //========= Security Descriptor =========

    private ResourceSecurityDescriptor securityDescriptor = null;

    public ResourceSecurityDescriptor getSecurityDescriptor() {
        if (this.securityDescriptor != null) {
            logger.debug("MJR security descriptor authz: "
                        + this.securityDescriptor.getAuthz());
        } else {
            logger.debug("MRJR security descriptor is null");
        }
        return this.securityDescriptor;
    }

    static {
        Util.registerTransport();
    }

    private static Log logger =
        LogFactory.getLog(ManagedMultiJobResource.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
    private static I18n coreI18n = I18n.getI18n(
        org.globus.wsrf.utils.Resources.class.getName());
}
