/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.service.factory.ManagedJobFactoryService;
import org.globus.exec.utils.Resources;
import org.globus.util.I18n;
import org.globus.wsrf.Constants;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.security.descriptor.SecurityDescriptor;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityConfig;

public class AuthorizationHelper
{
    static Log logger = LogFactory.getLog(AuthorizationHelper.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static final String GRIDMAP_AUTHZ_PDP_NAME = "gridmap";

    private static boolean authorizationGridmap = false;

    private AuthorizationHelper() { }

    public static void initialize()
    {
        logger.debug("getting factory service name");
        String factoryServiceName = ManagedJobFactoryService.getServicePath();

        try
        {
            logger.debug("getting security descriptor for JNDI path "
                        + factoryServiceName);
            SecurityDescriptor securityDescriptor
                = ServiceSecurityConfig.getSecurityDescriptor(
                    factoryServiceName);
            logger.debug("getting authz PDP name");
            String authz = securityDescriptor.getAuthz();
            String[] authzPDPNames = null;
            if (authz != null)
            {
                authzPDPNames = authz.split(" ");

                for (int index=0; index<authzPDPNames.length; index++)
                {
                    logger.debug("Detected authorization PDP plugin"
                               +" \"" + authzPDPNames[index] + "\".");

                    if (authzPDPNames[index].equals(GRIDMAP_AUTHZ_PDP_NAME))
                    {
                        logger.debug("Using gridmap authorization PDP plugin.");
                        AuthorizationHelper.authorizationGridmap = true;

                    }
                }
            }
        }
        catch (Exception e)
        {
            logger.debug(
                "Unable to determine authorization PDP plugin name.", e);
            throw new RuntimeException(
                "Unable to determine authorization PDP plugin name.", e);
        }
    }

    public static boolean isAuthorizationGridmap()
    {
        logger.debug("Entering/Exiting isAuthorizationGridmap()");
        return AuthorizationHelper.authorizationGridmap;
    }
}
