/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceContextException;
import org.globus.wsrf.Constants;

import org.globus.util.I18n;

import org.globus.exec.utils.Resources;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class ConfigurationHelper {

    public static ResourceHome getResourceHome(String service)
        throws NamingException
    {
        return (ResourceHome)getJNDIObject(service, Constants.HOME_NAME);
    }

    public static Object getResourceHomeConfiguration(String service)
        throws NamingException
    {
        return getJNDIObject(service, "homeConfiguration");
    }

    public static Object getResourceHomeConfiguration()
        throws NamingException, ResourceContextException
    {
        ResourceContext ctx = ResourceContext.getResourceContext();
        return getResourceHomeConfiguration(ctx.getService());
    }

    public static Object getServiceConfiguration(String service)
        throws NamingException
    {
        return getJNDIObject(service, "configuration");
    }

    public static Object getServiceConfiguration()
        throws NamingException, ResourceContextException
    {
        ResourceContext ctx = ResourceContext.getResourceContext();
        return getServiceConfiguration(ctx.getService());
    }

    public static Object getAuditDatabaseConfiguration(
        String                              service)
        throws                              NamingException
    {
        return getJNDIObject(service, "auditDatabaseConfiguration");
    }

    public static Object getAuditDatabaseConfiguration()
        throws                              NamingException,
                                            ResourceContextException
    {
        ResourceContext ctx = ResourceContext.getResourceContext();
        return getAuditDatabaseConfiguration(ctx.getService());
    }

    public static Object getJNDIObjectForCurrentService(String objectName)
        throws NamingException, ResourceContextException
    {
        ResourceContext ctx = ResourceContext.getResourceContext();
        return getJNDIObject(ctx.getService(), objectName);
    }


    /**
     *
     * @param service String The service within the JNDI context of which
     *                       the object is located
     * @param objectName String the name of the JNDI resource to look up.
     * @throws NamingException if the JNDI lookup fails
     * @return Object
     */
    public static Object getJNDIObject(String service, String objectName)
        throws NamingException
    {
        String jndiLookupString = null;
        Object jndiObject;
        try {
            Context initialContext = new InitialContext();
            if (!objectName.startsWith("/")) {
                objectName = "/" + objectName;
            }
            jndiLookupString = getServiceJNDILocation(service) + objectName;
            jndiObject = initialContext.lookup(jndiLookupString);
        } catch (NamingException e) {
            String errorMessage = i18n.getMessage(Resources.JNDI_CONFIG_ERROR,
                jndiLookupString);
            logger.error(errorMessage + "\n" + e.getMessage());
            throw e;
        }
        return jndiObject;
    }

    public static String getServiceJNDILocation(String service) {
        return Constants.JNDI_SERVICES_BASE_NAME + service;
    }

    public static String getServiceJNDILocation()
        throws ResourceContextException
    {
        ResourceContext ctx = ResourceContext.getResourceContext();
        return Constants.JNDI_SERVICES_BASE_NAME + ctx.getService();
    }

    private static Log logger =
        LogFactory.getLog(ConfigurationHelper.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
}
