/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import java.io.File;
import java.io.FileInputStream;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.axis.transport.local.LocalTransportUtils;

import org.globus.exec.service.factory.FactoryHomeConfiguration;
import org.globus.exec.service.factory.ManagedJobFactoryHome;
import org.globus.exec.utils.Resources;

import org.globus.util.I18n;

public class GlobusShToolsProperties extends Properties {
    private static Log logger =
        LogFactory.getLog(GlobusShToolsProperties.class.getName());
    private static String GLOBUS_SH_TOOLS_VARS
        = "libexec" + File.separator + "globus-sh-tools-vars.sh";

    private static GlobusShToolsProperties INSTANCE = null;

    private GlobusShToolsProperties() { }
    private GlobusShToolsProperties(Properties defaults) { }

    public static synchronized GlobusShToolsProperties getInstance()
    {
        if (logger.isDebugEnabled())
        {
            logger.debug("entering getInstance()");
        }

        if (INSTANCE == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.debug("INSTANCE is null. Creating new instance...");
            }
            INSTANCE = new GlobusShToolsProperties();

            FactoryHomeConfiguration factoryHomeConfiguration
                = ManagedJobFactoryHome.getConfiguration();
            File globusLocation
                = new File(factoryHomeConfiguration.getGlobusLocation());
            if (logger.isDebugEnabled())
            {
                logger.debug("globusLocation: " + globusLocation);
            }

            try {
                INSTANCE.load(new FileInputStream(
                    new File(globusLocation, GLOBUS_SH_TOOLS_VARS)));
            } catch (Exception e) {
                String errorMessage =
                    i18n.getMessage(Resources.GLOBUS_SH_TOOLS_NOT_FOUND);
                logger.error(errorMessage, e);
            }
        }

        if (logger.isDebugEnabled())
        {
            logger.debug("leaving getInstance()");
        }

        return INSTANCE;
    }

    public String getProperty(String key)
    {
        String propertyValue = null;
        String existingValue = super.getProperty(key);
        if (existingValue != null) {
            propertyValue = existingValue.replaceAll("\"", "");
        }
        else {
            String errorMessage =
                i18n.getMessage(Resources.GLOBUS_SH_PROPERTY_NOT_FOUND, key);
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
        return propertyValue;
    }

    public String getProperty(String key, String defaultValue)
    {
        return super.getProperty(key, defaultValue).replaceAll("\"", "");
    }

    private static I18n i18n = I18n.getI18n(Resources.class.getName());
}
