/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.utils;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.naming.InitialContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.message.MessageElement;

import org.globus.util.I18n;
import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.Topic;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.TopicListener;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.SimpleSubscriptionTopicListener;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.exec.utils.Resources;
import org.globus.exec.generated.TopicListenerDescriptionType;
import org.globus.exec.generated.TopicPathType;

public class SubscriptionPersistenceUtils
{
    /**
 * TODO: put in core in equivalent class
 */


    /**
     * Serialize any registered subscription topic listeners of type
     * SimpleSubscriptionListener.
     *
     * @param topicList The topic list on which the listeners are registered
     * @return TopicListenerDescriptionType[] the array of serializable
     *                      subscription data
     */
    public static TopicListenerDescriptionType[] getSerializableListeners(
        TopicList topicList)
        throws Exception
    {

        Map topicListenerTable =
            org.globus.wsrf.utils.SubscriptionPersistenceUtils.
                getListenerToTopicsTable(topicList);

        TopicListenerDescriptionType[] listenerArray =
            new TopicListenerDescriptionType[topicListenerTable.size()];
        Iterator topicListenerIterator =
            topicListenerTable.entrySet().iterator();

        int i = 0;
        while(topicListenerIterator.hasNext())
        {
            Map.Entry tableEntry = (Map.Entry) topicListenerIterator.next();
            SimpleSubscriptionTopicListener listener =
                (SimpleSubscriptionTopicListener) tableEntry.getKey();
            Iterator topicIterator = ((List) tableEntry.getValue()).iterator();

            TopicListenerDescriptionType listenerDescription =
                new TopicListenerDescriptionType();

            //set the subscription Home location
            listenerDescription.setSubscriptionHomeJNDILocation(
                listener.getSubscriptionHomeLocation());

            //set the subscription resource key
            ResourceKey key = listener.getSubscriptionResourceKey();
            AnyHelper.setAny(listenerDescription, key.toSOAPElement());

            //set the topics
            int topicPathCount = 0;
            List topicPaths = new ArrayList();
            while(topicIterator.hasNext())
            {
                List topicPathAsList = (List)topicIterator.next();
                TopicPathType topicPath = new TopicPathType();
                topicPath.setTopicPathElement(
                    (QName[])topicPathAsList.toArray(new QName[0]));
                topicPaths.add(topicPath);
                topicPathCount++;
            }
            TopicPathType[] topicPathArray =
                (TopicPathType[])topicPaths.toArray(new TopicPathType[0]);
            listenerDescription.setTopicPath(topicPathArray);

            listenerArray[i] = listenerDescription;

            i++;
        }

        return listenerArray;

    }

    /**
     * Add the deserialized listeners to the topics passed in parameter.
     *
     * @param topicList The topic list on which to register the deserialized
     *                  listeners
     */
    public static void addSubscriptionListenersToTopics(
        TopicListenerDescriptionType[] listenerDescriptions,
        TopicList topicList)
        throws Exception
    {
        if (listenerDescriptions == null) {
            logger.debug("array of deserialized subscriptions is null");
            return;
        }
        //for each listener decription,
        for (int i = 0; i < listenerDescriptions.length; i++) {
            //create the listener object

            String homeLocation =
                listenerDescriptions[i].getSubscriptionHomeJNDILocation();
            ResourceHome home = (ResourceHome)
                new InitialContext().lookup(homeLocation);
            Class keyClass = home.getKeyTypeClass();

            MessageElement keyAnyElement = listenerDescriptions[i].get_any()[0];
            ResourceKey key = new SimpleResourceKey(
                keyAnyElement,
                keyClass);

            SimpleSubscriptionTopicListener listener =
                new SimpleSubscriptionTopicListener(key, homeLocation);

            //register the listener with each of its topics
            TopicPathType[] topics = listenerDescriptions[i].getTopicPath();
            for (int j = 0; j < topics.length; j++) {
                //make topic path as List
                List topicPath = Arrays.asList(topics[j].getTopicPathElement());
                Topic topic = topicList.getTopic(topicPath);
                topic.addTopicListener(listener);
                if (logger.isDebugEnabled()) {
                    logger.debug("Registered listener of subscription key " +
                                 key.getValue() +  " with topic starting with "+
                                 topicPath.get(0));
                }
            } //end for
        } //end for
    }

    private static Log logger = LogFactory.getLog(
        SubscriptionPersistenceUtils.class.getName());

    private static I18n i18n = I18n.getI18n(Resources.class.getName());

}

